from unittest.mock import Mock

import numpy as np
import pytest

from napari.layers import Shapes
from napari.layers.base._base_constants import ActionType
from napari.layers.shapes.shapes import Mode
from napari.settings import get_settings
from napari.utils._test_utils import read_only_mouse_event
from napari.utils.interactions import (
    mouse_double_click_callbacks,
    mouse_move_callbacks,
    mouse_press_callbacks,
    mouse_release_callbacks,
)


@pytest.fixture
def create_known_shapes_layer():
    """Create shapes layer with known coordinates

    Returns
    -------
    layer : napari.layers.Shapes
        Shapes layer.
    n_shapes : int
        Number of shapes in the shapes layer
    known_non_shape : list
        Data coordinates that are known to contain no shapes. Useful during
        testing when needing to guarantee no shape is clicked on.
    """
    data = [[[1, 3], [8, 4]], [[10, 10], [15, 4]]]
    known_non_shape = [20, 30]
    n_shapes = len(data)

    layer = Shapes(data)
    # very zoomed in, guaranteed no overlap between vertices
    layer.scale_factor = 0.001
    assert layer.ndim == 2
    assert len(layer.data) == n_shapes
    assert len(layer.selected_data) == 0

    return layer, n_shapes, known_non_shape


def test_not_adding_or_selecting_shape(create_known_shapes_layer):
    """Don't add or select a shape by clicking on one in pan_zoom mode."""
    layer, n_shapes, _ = create_known_shapes_layer
    layer.mode = 'pan_zoom'

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
    )
    mouse_release_callbacks(layer, event)

    # Check no new shape added and non selected
    assert len(layer.data) == n_shapes
    assert len(layer.selected_data) == 0


@pytest.mark.parametrize('shape_type', ['rectangle', 'ellipse', 'line'])
def test_add_simple_shape(shape_type, create_known_shapes_layer):
    """Add simple shape by clicking in add mode."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    # Add shape at location where non exists
    layer.mode = 'add_' + shape_type

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=known_non_shape,
    )
    mouse_press_callbacks(layer, event)

    known_non_shape_end = [40, 60]
    # Simulate drag end
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=known_non_shape_end,
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        position=known_non_shape_end,
    )
    mouse_release_callbacks(layer, event)

    # Check new shape added at coordinates
    assert len(layer.data) == n_shapes + 1
    np.testing.assert_allclose(layer.data[-1][0], known_non_shape)
    new_shape_max = np.max(layer.data[-1], axis=0)
    np.testing.assert_allclose(new_shape_max, known_non_shape_end)
    assert layer.shape_type[-1] == shape_type

    # Ensure it's selected, accounting for zero-indexing
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {n_shapes}


def test_line_fixed_angles(create_known_shapes_layer):
    """Draw line with fixed angles."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    layer.mode = 'add_line'

    # set _fixed_aspect, like Shift would
    layer._fixed_aspect = True

    # Simulate click with shift
    event = read_only_mouse_event(
        type='mouse_press',
        position=known_non_shape,
    )
    mouse_press_callbacks(layer, event)

    known_non_shape_end = [40, 60]
    # Simulate drag with shift
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=known_non_shape_end,
    )
    mouse_move_callbacks(layer, event)

    # Simulate release with shift
    event = read_only_mouse_event(
        type='mouse_release',
        position=known_non_shape_end,
    )
    mouse_release_callbacks(layer, event)

    new_line = layer.data[-1][-1] - layer.data[-1][0]

    # Check new shape added at coordinates
    assert len(layer.data) == n_shapes + 1
    # start should match mouse event
    assert np.allclose(
        layer.data[-1][0], np.asarray(known_non_shape).astype(float)
    )
    # With _fixed_aspect, the line end won't be at the end event
    assert not np.allclose(
        layer.data[-1][-1], np.asarray(known_non_shape_end).astype(float)
    )
    # with _fixed_aspect the angle of the line should be 45 degrees
    theta = np.degrees(np.arctan2(new_line[1], new_line[0]))
    assert np.allclose(theta, 45.0)


def test_path_tablet(create_known_shapes_layer):
    layer, n_shapes, known_non_shape = create_known_shapes_layer
    desired_shape = np.array([[20, 30], [10, 50], [60, 40], [80, 20]])

    layer.mode = 'add_path'
    event = read_only_mouse_event(
        type='mouse_press',
        is_dragging=True,
        position=desired_shape[0],
        pos=desired_shape[0],
    )
    mouse_press_callbacks(layer, event)
    assert layer.shape_type[-1] == 'path'

    for coord in desired_shape[1:]:
        event = read_only_mouse_event(
            type='mouse_move',
            is_dragging=True,
            position=coord,
            pos=coord,
        )
        mouse_move_callbacks(layer, event)

    event = read_only_mouse_event(
        type='mouse_release',
        is_dragging=True,
        position=desired_shape[-1],
        pos=desired_shape[-1],
    )
    mouse_release_callbacks(layer, event)

    assert len(layer.data) == n_shapes + 1
    assert np.array_equal(desired_shape, layer.data[-1])

    assert layer.shape_type[-1] == 'path'
    assert not layer._is_creating

    # Ensure it's selected, accounting for zero-indexing
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {n_shapes}


def test_polyline_mouse(create_known_shapes_layer):
    layer, n_shapes, known_non_shape = create_known_shapes_layer
    desired_shape = np.array([[20, 30], [10, 50], [60, 40], [80, 20]])

    layer.mode = 'add_path'

    event = read_only_mouse_event(
        type='mouse_press',
        position=desired_shape[0],
        pos=desired_shape[0],
    )
    mouse_press_callbacks(layer, event)
    assert layer.shape_type[-1] == 'path'

    for coord in desired_shape[1:]:
        event = read_only_mouse_event(
            type='mouse_move',
            position=coord,
            pos=coord,
        )
        mouse_move_callbacks(layer, event)

    event = read_only_mouse_event(
        type='mouse_press',
        position=desired_shape[-1],
        pos=desired_shape[-1],
    )
    mouse_press_callbacks(layer, event)

    assert len(layer.data) == n_shapes + 1
    assert np.array_equal(desired_shape, layer.data[-1])
    assert layer.shape_type[-1] == 'path'
    assert not layer._is_creating

    # Ensure it's selected, accounting for zero-indexing
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {n_shapes}


def test_polygon_lasso_tablet(create_known_shapes_layer):
    """Draw polygon with tablet simulated by mouse drag event."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer
    desired_shape = np.array([[20, 30], [10, 50], [60, 40], [80, 20]])

    get_settings().experimental.rdp_epsilon = 0
    layer.mode = 'add_polygon_lasso'

    event = read_only_mouse_event(
        type='mouse_press',
        is_dragging=True,
        position=desired_shape[0],
        pos=desired_shape[0],
    )
    mouse_press_callbacks(layer, event)

    assert layer.shape_type[-1] != 'polygon'

    for coord in desired_shape[1:]:
        event = read_only_mouse_event(
            type='mouse_move',
            is_dragging=True,
            position=coord,
            pos=coord,
        )
        mouse_move_callbacks(layer, event)

    event = read_only_mouse_event(
        type='mouse_release',
        is_dragging=True,
        position=desired_shape[-1],
        pos=desired_shape[-1],
    )
    mouse_release_callbacks(layer, event)

    assert len(layer.data) == n_shapes + 1
    assert np.array_equal(desired_shape, layer.data[-1])
    assert layer.shape_type[-1] == 'polygon'
    assert not layer._is_creating

    # Ensure it's selected, accounting for zero-indexing
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {n_shapes}


def test_polygon_lasso_mouse(create_known_shapes_layer):
    """Draw polygon with mouse. Events in sequence are mouse press, release, move, press, release"""
    layer, n_shapes, known_non_shape = create_known_shapes_layer
    desired_shape = np.array([[20, 30], [10, 50], [60, 40], [80, 20]])

    get_settings().experimental.rdp_epsilon = 0
    layer.mode = 'add_polygon_lasso'

    event = read_only_mouse_event(
        type='mouse_press',
        position=desired_shape[0],
        pos=desired_shape[0],
    )
    mouse_press_callbacks(layer, event)
    assert layer.shape_type[-1] != 'polygon'

    for coord in desired_shape[1:]:
        event = read_only_mouse_event(
            type='mouse_move',
            position=coord,
            pos=coord,
        )
        mouse_move_callbacks(layer, event)

    event = read_only_mouse_event(
        type='mouse_press',
        position=desired_shape[-1],
        pos=desired_shape[-1],
    )
    mouse_press_callbacks(layer, event)

    assert len(layer.data) == n_shapes + 1
    assert np.array_equal(desired_shape, layer.data[-1])
    assert layer.shape_type[-1] == 'polygon'
    assert not layer._is_creating

    # Ensure it's selected, accounting for zero-indexing
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {n_shapes}


def test_distance_polygon_creating(create_known_shapes_layer):
    """Test that distance threshold in polygon creating works as intended"""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    # While drawing only 2 of the vertices should be added to shape data because distance threshold is 10
    vertices = [[x, 0] for x in range(11)]
    layer.mode = 'add_polygon_lasso'

    event = read_only_mouse_event(
        type='mouse_press',
        position=vertices[0],
        pos=vertices[0],
    )
    mouse_press_callbacks(layer, event)
    for coord in vertices[1:]:
        event = read_only_mouse_event(
            type='mouse_move',
            position=coord,
            pos=coord,
        )
        mouse_move_callbacks(layer, event)

    assert len(layer.data[-1] == 2)


@pytest.mark.parametrize('shape_type', ['polyline', 'polygon'])
def test_add_complex_shape(shape_type, create_known_shapes_layer):
    """Add simple shape by clicking in add mode."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    desired_shape = [[20, 30], [10, 50], [60, 40], [80, 20]]
    # Add shape at location where non exists
    layer.mode = 'add_' + shape_type

    for coord in desired_shape:
        # Simulate move, click, and release
        event = read_only_mouse_event(
            type='mouse_move',
            position=coord,
            pos=np.array(coord, dtype=float),
        )
        mouse_move_callbacks(layer, event)
        event = read_only_mouse_event(
            type='mouse_press',
            position=coord,
            pos=np.array(coord, dtype=float),
        )
        mouse_press_callbacks(layer, event)
        event = read_only_mouse_event(
            type='mouse_release',
            position=coord,
            pos=np.array(coord, dtype=float),
        )
        mouse_release_callbacks(layer, event)

    # finish drawing
    end_click = read_only_mouse_event(
        type='mouse_double_click',
        position=coord,
    )
    assert layer.mouse_double_click_callbacks
    mouse_double_click_callbacks(layer, end_click)

    # Check new shape added at coordinates
    assert len(layer.data) == n_shapes + 1
    assert layer.data[-1].shape, desired_shape.shape
    np.testing.assert_allclose(layer.data[-1], desired_shape)
    shape_type = shape_type if shape_type == 'polygon' else 'path'
    assert layer.shape_type[-1] == shape_type

    # Ensure it's selected, accounting for zero-indexing
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {n_shapes}


@pytest.mark.parametrize(
    'shape_type_vertices',
    [
        ('polyline', [[20, 30], [20, 30]]),
        ('polygon', [[20, 30], [10, 50], [10, 50]]),
    ],
)
def test_add_invalid_shape(shape_type_vertices, create_known_shapes_layer):
    """Check invalid shape clicking behavior in add polygon mode."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    # Add shape at location where non exists
    shape_type, shape_vertices = shape_type_vertices
    layer.mode = f'add_{shape_type}'

    for coord in shape_vertices:
        # Simulate move, click, and release
        event = read_only_mouse_event(
            type='mouse_move',
            position=coord,
            pos=np.array(coord, dtype=float),
        )
        mouse_move_callbacks(layer, event)
        event = read_only_mouse_event(
            type='mouse_press',
            position=coord,
            pos=np.array(coord, dtype=float),
        )
        mouse_press_callbacks(layer, event)
        event = read_only_mouse_event(
            type='mouse_release',
            position=coord,
            pos=np.array(coord, dtype=float),
        )
        mouse_release_callbacks(layer, event)

    # Although the shape/polygon being created is invalid, three shapes should
    # be available until it is marked as finished via mouse double click
    assert len(layer.data) == n_shapes + 1

    # finish drawing causing the removal of the in progress shape since is invalid
    end_click = read_only_mouse_event(
        type='mouse_double_click',
        position=coord,
        pos=np.array(coord, dtype=float),
    )
    assert layer.mouse_double_click_callbacks
    mouse_double_click_callbacks(layer, end_click)

    # Ensure no data is selected, number of shapes is the initial one and
    # last cursor position variable was reset
    assert len(layer.selected_data) == 0
    assert len(layer.data) == n_shapes
    assert layer._last_cursor_position is None


def test_vertex_insert(create_known_shapes_layer):
    """Add vertex to shape."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    layer.events.data = Mock()
    n_coord = len(layer.data[0])
    layer.mode = 'vertex_insert'
    layer.selected_data = {0}
    old_data = layer.data
    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=known_non_shape,
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag end
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=known_non_shape,
    )
    mouse_move_callbacks(layer, event)

    # Check new shape added at coordinates
    assert len(layer.data) == n_shapes
    assert len(layer.data[0]) == n_coord + 1
    assert layer.events.data.call_args_list[0][1] == {
        'value': old_data,
        'action': ActionType.CHANGING,
        'data_indices': tuple(layer.selected_data),
        'vertex_indices': ((2,),),
    }
    assert layer.events.data.call_args[1] == {
        'value': layer.data,
        'action': ActionType.CHANGED,
        'data_indices': tuple(layer.selected_data),
        'vertex_indices': ((2,),),
    }
    np.testing.assert_allclose(
        np.min(abs(layer.data[0] - known_non_shape), axis=0), [0, 0]
    )


def test_vertex_remove(create_known_shapes_layer):
    """Remove vertex from shape."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer
    old_data = layer.data
    layer.events.data = Mock()
    n_coord = len(layer.data[0])
    layer.mode = 'vertex_remove'
    select = {0}
    layer.selected_data = select
    position = tuple(layer.data[0][0])

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=position,
    )
    mouse_press_callbacks(layer, event)
    assert layer.events.data.call_args_list[0][1] == {
        'value': old_data,
        'action': ActionType.CHANGING,
        'data_indices': tuple(
            select,
        ),
        'vertex_indices': ((0,),),
    }
    assert layer.events.data.call_args[1] == {
        'value': layer.data,
        'action': ActionType.CHANGED,
        'data_indices': tuple(
            select,
        ),
        'vertex_indices': ((0,),),
    }
    assert len(layer.data) == n_shapes
    assert len(layer.data[0]) == n_coord - 1


@pytest.mark.parametrize('mode', ['select', 'direct'])
def test_select_shape(mode, create_known_shapes_layer):
    """Select a shape by clicking on one in select mode."""
    layer, n_shapes, _ = create_known_shapes_layer

    layer.mode = mode
    position = tuple(layer.data[0][0])

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=position,
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        position=position,
    )
    mouse_release_callbacks(layer, event)

    # Check clicked shape selected
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {0}


def test_drag_shape(create_known_shapes_layer):
    """Select and drag vertex."""
    layer, n_shapes, _ = create_known_shapes_layer
    layer.events.data = Mock()

    old_data = layer.data
    layer.mode = 'select'
    orig_data = layer.data[0].copy()
    assert len(layer.selected_data) == 0

    position = tuple(np.mean(layer.data[0], axis=0))

    # Check shape under cursor
    value = layer.get_value(position, world=True)
    assert value == (0, None)

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=position,
    )
    mouse_press_callbacks(layer, event)
    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        position=position,
    )
    mouse_release_callbacks(layer, event)

    assert len(layer.selected_data) == 1
    assert layer.selected_data == {0}

    # Check shape but not vertex under cursor
    value = layer.get_value(event.position, world=True)
    assert value == (0, None)

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        is_dragging=True,
        position=position,
    )
    mouse_press_callbacks(layer, event)
    # start drag event
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=position,
    )
    mouse_move_callbacks(layer, event)
    position = tuple(np.add(position, [10, 5]))
    # Simulate move, click, and release
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=position,
    )
    mouse_move_callbacks(layer, event)
    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        is_dragging=True,
        position=position,
    )
    mouse_release_callbacks(layer, event)

    # Check clicked shape selected
    vertex_indices = (tuple(range(len(layer.data[0]))),)
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {0}
    assert layer.events.data.call_args_list[0][1] == {
        'value': old_data,
        'action': ActionType.CHANGING,
        'data_indices': (0,),
        'vertex_indices': vertex_indices,
    }
    assert layer.events.data.call_args[1] == {
        'value': layer.data,
        'action': ActionType.CHANGED,
        'data_indices': (0,),
        'vertex_indices': vertex_indices,
    }
    np.testing.assert_allclose(layer.data[0], orig_data + np.array([10, 5]))


def test_rotate_shape(create_known_shapes_layer):
    """Select and drag handle to rotate shape."""
    layer, n_shapes, _ = create_known_shapes_layer

    layer.mode = 'select'
    layer.selected_data = {1}
    # get the position of the rotation handle
    position = tuple(layer._selected_box[9])
    # get the vertexes
    original_data = layer.data[1].copy()

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        is_dragging=True,
        position=position,
    )
    mouse_press_callbacks(layer, event)
    # start drag event
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=position,
    )
    mouse_move_callbacks(layer, event)

    # drag in the handle to bottom midpoint vertex to rotate 180 degrees
    position = tuple(layer._selected_box[3])
    # Simulate move, click, and release
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=position,
    )
    mouse_move_callbacks(layer, event)
    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        is_dragging=True,
        position=position,
    )
    mouse_release_callbacks(layer, event)

    # Check shape was rotated
    np.testing.assert_allclose(layer.data[1][2], original_data[0])


def test_drag_vertex(create_known_shapes_layer):
    """Select and drag vertex."""
    layer, n_shapes, _ = create_known_shapes_layer
    layer.events.data = Mock()
    layer.mode = 'direct'
    layer.selected_data = {0}
    old_position = tuple(layer.data[0][0])

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=old_position,
    )
    mouse_press_callbacks(layer, event)

    new_position = [0, 0]
    assert np.all(new_position != old_position)

    # Simulate move, click, and release
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=new_position,
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        is_dragging=True,
        position=new_position,
    )
    mouse_release_callbacks(layer, event)

    # Check clicked shape selected
    vertex_indices = (tuple(range(len(layer.data[0]))),)
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {0}
    assert layer.events.data.call_args[1] == {
        'value': layer.data,
        'action': ActionType.CHANGED,
        'data_indices': (0,),
        'vertex_indices': vertex_indices,
    }
    np.testing.assert_allclose(layer.data[0][0], [0, 0])


@pytest.mark.parametrize(
    'mode',
    [
        'select',
        'direct',
        'add_rectangle',
        'add_ellipse',
        'add_line',
        'add_polygon',
        'add_path',
        'vertex_insert',
        'vertex_remove',
    ],
)
def test_after_in_add_mode_shape(mode, create_known_shapes_layer):
    """Don't add or select a shape by clicking on one in pan_zoom mode."""
    layer, n_shapes, _ = create_known_shapes_layer

    layer.mode = mode
    layer.mode = 'pan_zoom'
    position = tuple(layer.data[0][0])

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=position,
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        position=position,
    )
    mouse_release_callbacks(layer, event)

    # Check no new shape added and non selected
    assert len(layer.data) == n_shapes
    assert len(layer.selected_data) == 0


@pytest.mark.parametrize(
    'mode',
    [
        'add_polygon',
        'add_path',
    ],
)
def test_clicking_the_same_point_is_not_crashing(
    mode, create_known_shapes_layer
):
    layer, n_shapes, _ = create_known_shapes_layer

    layer.mode = mode
    position = tuple(layer.data[0][0])

    for _ in range(2):
        event = read_only_mouse_event(type='mouse_press', position=position)
        mouse_press_callbacks(layer, event)

        event = read_only_mouse_event(type='mouse_release', position=position)
        mouse_release_callbacks(layer, event)

    # If there was no move event between the two clicks, we expect value[1] must be None
    assert layer.get_value(event.position, world=True)[1] is None


@pytest.mark.parametrize(
    'mode',
    [
        'add_polygon',
        'add_polyline',
    ],
)
def test_is_creating_is_false_on_creation(mode, create_known_shapes_layer):
    layer, n_shapes, _ = create_known_shapes_layer

    layer.mode = mode
    position = tuple(layer.data[0][0])

    def is_creating_is_True(event):
        assert event.source._is_creating

    def is_creating_is_False(event):
        assert not event.source._is_creating

    assert not layer._is_creating
    layer.events.set_data.connect(is_creating_is_True)

    event = read_only_mouse_event(type='mouse_press', position=position)
    mouse_press_callbacks(layer, event)

    assert layer._is_creating

    event = read_only_mouse_event(type='mouse_release', position=position)
    mouse_release_callbacks(layer, event)

    assert layer._is_creating

    layer.events.set_data.disconnect(is_creating_is_True)
    layer.events.set_data.connect(is_creating_is_False)
    end_click = read_only_mouse_event(
        type='mouse_double_click',
        position=position,
    )
    mouse_double_click_callbacks(layer, end_click)

    assert not layer._is_creating


@pytest.mark.parametrize('mode', ['select', 'direct'])
def test_unselect_select_shape(mode, create_known_shapes_layer):
    """Select a shape by clicking on one in select mode."""
    layer, n_shapes, _ = create_known_shapes_layer

    layer.mode = mode
    position = tuple(layer.data[0][0])
    layer.selected_data = {1}

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=position,
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        position=position,
    )
    mouse_release_callbacks(layer, event)

    # Check clicked shape selected
    assert len(layer.selected_data) == 1
    assert layer.selected_data == {0}


@pytest.mark.parametrize('mode', ['select', 'direct'])
def test_not_selecting_shape(mode, create_known_shapes_layer):
    """Don't select a shape by not clicking on one in select mode."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    layer.mode = mode

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=known_non_shape,
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        position=known_non_shape,
    )
    mouse_release_callbacks(layer, event)

    # Check clicked shape selected
    assert len(layer.selected_data) == 0


@pytest.mark.parametrize('mode', ['select', 'direct'])
def test_unselecting_shapes(mode, create_known_shapes_layer):
    """Unselect shapes by not clicking on one in select mode."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    layer.mode = mode
    layer.selected_data = {0, 1}
    assert len(layer.selected_data) == 2

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=known_non_shape,
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        position=known_non_shape,
    )
    mouse_release_callbacks(layer, event)

    # Check clicked shape selected
    assert len(layer.selected_data) == 0


@pytest.mark.parametrize('mode', ['select', 'direct'])
def test_selecting_shapes_with_drag(mode, create_known_shapes_layer):
    """Select all shapes when drag box includes all of them."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    layer.mode = mode

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=known_non_shape,
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag start
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=known_non_shape,
    )
    mouse_move_callbacks(layer, event)

    # Simulate drag end
    event = read_only_mouse_event(type='mouse_move', is_dragging=True)
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        is_dragging=True,
    )
    mouse_release_callbacks(layer, event)

    # Check all shapes selected as drag box contains them
    assert len(layer.selected_data) == n_shapes


@pytest.mark.parametrize('mode', ['select', 'direct'])
def test_selecting_no_shapes_with_drag(mode, create_known_shapes_layer):
    """Select all shapes when drag box includes all of them."""
    layer, n_shapes, known_non_shape = create_known_shapes_layer

    layer.mode = mode

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        position=known_non_shape,
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag start
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=known_non_shape,
    )
    mouse_move_callbacks(layer, event)

    # Simulate drag end
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=(50, 60),
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        is_dragging=True,
        position=(50, 60),
    )
    mouse_release_callbacks(layer, event)

    # Check no shapes selected as drag box doesn't contain them
    assert len(layer.selected_data) == 0


@pytest.mark.parametrize(
    'attr', ['_move_modes', '_drag_modes', '_cursor_modes']
)
def test_all_modes_covered(attr):
    """
    Test that all dictionaries modes have all the keys, this simplify the handling logic
    As we do not need to test whether a key is in a dict or not.
    """
    mode_dict = getattr(Shapes, attr)
    assert {k.value for k in mode_dict} == set(Mode.keys())


@pytest.mark.parametrize(
    ('pre_selection', 'on_point', 'modifier'),
    [
        (set(), True, []),
        ({1}, True, []),
    ],
)
def test_drag_start_selection(
    create_known_shapes_layer, pre_selection, on_point, modifier
):
    """Check layer drag start and drag box behave as expected."""
    layer, n_points, known_non_point = create_known_shapes_layer
    layer.mode = 'select'
    layer.selected_data = pre_selection

    if on_point:
        initial_position = tuple(layer.data[0].mean(axis=0))
    else:
        initial_position = tuple(known_non_point)
    zero_pos = [0, 0]

    value = layer.get_value(initial_position, world=True)
    assert value[0] == 0
    assert layer._drag_start is None
    assert layer._drag_box is None
    assert layer.selected_data == pre_selection

    # Simulate click
    event = read_only_mouse_event(
        type='mouse_press',
        is_dragging=True,
        modifiers=modifier,
        position=initial_position,
    )
    mouse_press_callbacks(layer, event)

    if modifier:
        if not on_point:
            assert layer.selected_data == pre_selection
        elif 0 in pre_selection:
            assert layer.selected_data == pre_selection - {0}
        else:
            assert layer.selected_data == pre_selection | {0}
    elif not on_point:
        assert layer.selected_data == set()
    elif 0 in pre_selection:
        assert layer.selected_data == pre_selection
    else:
        assert layer.selected_data == {0}

    if len(layer.selected_data) > 0:
        center_list = []
        for idx in layer.selected_data:
            center_list.append(layer.data[idx].mean(axis=0))
        center = np.mean(center_list, axis=0)
    else:
        center = [0, 0]

    if not modifier:
        start_position = [
            initial_position[0] - center[0],
            initial_position[1] - center[1],
        ]
    else:
        start_position = initial_position

    is_point_move = len(layer.selected_data) > 0 and on_point and not modifier

    np.testing.assert_array_equal(layer._drag_start, start_position)

    # Simulate drag start on a different position
    offset_position = [initial_position[0] + 20, initial_position[1] + 20]
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=offset_position,
        modifiers=modifier,
    )
    mouse_move_callbacks(layer, event)

    # Initial mouse_move is already considered a move and not a press.
    # Therefore, the _drag_start value should be identical and the data or drag_box should reflect
    # the mouse position.
    np.testing.assert_array_equal(layer._drag_start, start_position)
    if is_point_move:
        if 0 in layer.selected_data:
            np.testing.assert_array_equal(
                layer.data[0].mean(axis=0),
                [offset_position[0], offset_position[1]],
            )
        else:
            raise AssertionError('Unreachable code')  # pragma: no cover
    else:
        np.testing.assert_array_equal(
            layer._drag_box, [initial_position, offset_position]
        )

    # Simulate drag start on new different position
    offset_position = zero_pos
    event = read_only_mouse_event(
        type='mouse_move',
        is_dragging=True,
        position=offset_position,
        modifiers=modifier,
    )
    mouse_move_callbacks(layer, event)

    # Initial mouse_move is already considered a move and not a press.
    # Therefore, the _drag_start value should be identical and the data or drag_box should reflect
    # the mouse position.
    np.testing.assert_array_equal(layer._drag_start, start_position)
    if is_point_move:
        if 0 in layer.selected_data:
            np.testing.assert_array_equal(
                layer.data[0].mean(axis=0),
                [offset_position[0], offset_position[1]],
            )
        else:
            raise AssertionError('Unreachable code')  # pragma: no cover
    else:
        np.testing.assert_array_equal(
            layer._drag_box, [initial_position, offset_position]
        )

    # Simulate release
    event = read_only_mouse_event(
        type='mouse_release',
        is_dragging=True,
        modifiers=modifier,
        position=offset_position,
    )
    mouse_release_callbacks(layer, event)

    if on_point and 0 in pre_selection and modifier:
        assert layer.selected_data == pre_selection - {0}
    elif on_point and 0 in pre_selection and not modifier:
        assert layer.selected_data == pre_selection
    elif on_point and 0 not in pre_selection and modifier:
        assert layer.selected_data == pre_selection | {0}
    elif on_point and 0 not in pre_selection and not modifier:
        assert layer.selected_data == {0}
    elif 0 in pre_selection and modifier:
        assert 0 not in layer.selected_data
        assert layer.selected_data == (set(range(n_points)) - pre_selection)
    elif 0 in pre_selection and not modifier:
        assert 0 in layer.selected_data
        assert layer.selected_data == set(range(n_points))
    elif 0 not in pre_selection and modifier:
        assert 0 in layer.selected_data
        assert layer.selected_data == (set(range(n_points)) - pre_selection)
    elif 0 not in pre_selection and not modifier:
        assert 0 in layer.selected_data
        assert layer.selected_data == set(range(n_points))
    else:
        pytest.fail('Unreachable code')
    assert layer._drag_box is None
    assert layer._drag_start is None



import numpy as np


def calculate_barycentric_coordinates(
    point: np.ndarray, triangle_vertices: np.ndarray
) -> np.ndarray:
    """Calculate the barycentric coordinates for a point in a triangle.

    http://gamedev.stackexchange.com/questions/23743/whats-the-most-efficient-way-to-find-barycentric-coordinates

    Parameters
    ----------
    point : np.ndarray
        The coordinates of the point for which to calculate the barycentric coordinate.
    triangle_vertices : np.ndarray
        (3, D) array containing the triangle vertices.

    Returns
    -------
    barycentric_coorinates : np.ndarray
        The barycentric coordinate [u, v, w], where u, v, and w are the
        barycentric coordinates for the first, second, third triangle
        vertex, respectively.
    """
    vertex_a = triangle_vertices[0, :]
    vertex_b = triangle_vertices[1, :]
    vertex_c = triangle_vertices[2, :]

    v0 = vertex_b - vertex_a
    v1 = vertex_c - vertex_a
    v2 = point - vertex_a
    d00 = np.dot(v0, v0)
    d01 = np.dot(v0, v1)
    d11 = np.dot(v1, v1)
    d20 = np.dot(v2, v0)
    d21 = np.dot(v2, v1)
    denominator = d00 * d11 - d01 * d01
    v = (d11 * d20 - d01 * d21) / denominator
    w = (d00 * d21 - d01 * d20) / denominator
    u = 1 - v - w
    return np.array([u, v, w])



from napari.types import PathLike


class MultipleReaderError(RuntimeError):
    """Multiple readers are available for paths and none explicitly chosen.

    Thrown when the viewer model tries to open files but multiple
    reader plugins are available that could claim them. User must
    make an explicit choice out of the available readers before opening
    files.

    Parameters
    ----------
    message: str
        error description
    available_readers : List[str]
        list of available reader plugins for path
    paths: List[str]
        file paths for reading

    Attributes
    ----------
    message: str
        error description
    available_readers : List[str]
        list of available reader plugins for path
    paths: List[str]
        file paths for reading
    """

    def __init__(
        self,
        message: str,
        available_readers: list[str],
        paths: list[PathLike],
        *args: object,
    ) -> None:
        super().__init__(message, *args)
        self.available_plugins = available_readers
        self.paths = paths


class ReaderPluginError(ValueError):
    """A reader plugin failed while trying to open paths.

    This error is thrown either when the only available plugin
    failed to read the paths, or when the plugin associated with the
    paths' file extension failed.

    Parameters
    ----------
    message: str
        error description
    reader_plugin : str
        plugin that was tried
    paths: List[str]
        file paths for reading

    Attributes
    ----------
    message: str
        error description
    reader_plugin : str
        plugin that was tried
    paths: List[str]
        file paths for reading
    """

    def __init__(
        self,
        message: str,
        reader_plugin: str,
        paths: list[PathLike],
        *args: object,
    ) -> None:
        super().__init__(message, *args)
        self.reader_plugin = reader_plugin
        self.paths = paths


class NoAvailableReaderError(ValueError):
    """No reader plugins are available to open the chosen file

    Parameters
    ----------
    message: str
        error description
    paths: List[str]
        file paths for reading

    Attributes
    ----------
    message: str
        error description
    paths: List[str]
        file paths for reading
    """

    def __init__(
        self, message: str, paths: list[PathLike], *args: object
    ) -> None:
        super().__init__(message, *args)
        self.paths = paths



import numpy as np


def dims_scroll(viewer, event):
    """Scroll the dimensions slider."""
    if 'Control' not in event.modifiers:
        return
    if event.native.inverted():
        viewer.dims._scroll_progress += event.delta[1]
    else:
        viewer.dims._scroll_progress -= event.delta[1]
    while abs(viewer.dims._scroll_progress) >= 1:
        if viewer.dims._scroll_progress < 0:
            viewer.dims._increment_dims_left()
            viewer.dims._scroll_progress += 1
        else:
            viewer.dims._increment_dims_right()
            viewer.dims._scroll_progress -= 1


def double_click_to_zoom(viewer, event):
    """Zoom in on double click by zoom_factor; zoom out with Alt."""
    if (
        viewer.layers.selection.active
        and viewer.layers.selection.active.mode != 'pan_zoom'
    ):
        return
    # if Alt held down, zoom out instead
    zoom_factor = 0.5 if 'Alt' in event.modifiers else 2
    viewer.camera.zoom *= zoom_factor
    if viewer.dims.ndisplay == 3 and viewer.dims.ndim == 3:
        viewer.camera.center = np.asarray(viewer.camera.center) + (
            np.asarray(event.position)[np.asarray(viewer.dims.displayed)]
            - np.asarray(viewer.camera.center)
        ) * (1 - 1 / zoom_factor)
    else:
        viewer.camera.center = np.asarray(viewer.camera.center)[-2:] + (
            np.asarray(event.position)[-2:]
            - np.asarray(viewer.camera.center)[-2:]
        ) * (1 - 1 / zoom_factor)



import pytest
from npe2 import PluginManager, PluginManifest, __version__ as npe2_version
from npe2.manifest.schema import ContributionPoints
from packaging.version import parse as parse_version

from napari._pydantic_compat import ValidationError
from napari.resources._icons import PLUGIN_FILE_NAME
from napari.settings import get_settings
from napari.utils.theme import (
    Theme,
    _install_npe2_themes,
    available_themes,
    get_theme,
    is_theme_available,
    register_theme,
    unregister_theme,
)


def test_default_themes():
    themes = available_themes()
    assert 'dark' in themes
    assert 'light' in themes
    assert 'system' in themes


def test_get_theme():
    # get theme in the old-style dict format
    theme = get_theme('dark').to_rgb_dict()
    assert isinstance(theme, dict)

    # get theme in the new model-based format
    theme = get_theme('dark')
    assert isinstance(theme, Theme)


def test_get_system_theme(monkeypatch):
    monkeypatch.setattr('napari.utils.theme.get_system_theme', lambda: 'light')
    theme = get_theme('system')
    # should return the theme specified by get_system_theme
    assert theme.id == 'light'


def test_register_theme():
    # Check that blue theme is not listed in available themes
    themes = available_themes()
    assert 'test_blue' not in themes

    # Create new blue theme based on dark theme
    blue_theme = get_theme('dark').to_rgb_dict()
    blue_theme.update(
        background='rgb(28, 31, 48)',
        foreground='rgb(45, 52, 71)',
        primary='rgb(80, 88, 108)',
        current='rgb(184, 112, 0)',
    )

    # Register blue theme
    register_theme('test_blue', blue_theme, 'test')

    # Check that blue theme is listed in available themes
    themes = available_themes()
    assert 'test_blue' in themes

    # Check that the dark theme has not been overwritten
    dark_theme = get_theme('dark').to_rgb_dict()
    assert dark_theme['background'] != blue_theme['background']

    # Check that blue theme can be gotten from available themes
    theme = get_theme('test_blue').to_rgb_dict()
    assert theme['background'] == blue_theme['background']

    theme = get_theme('test_blue')
    assert theme.background.as_rgb() == blue_theme['background']


def test_unregister_theme():
    # Create new blue theme based on dark theme
    blue_theme = get_theme('dark').to_rgb_dict()
    blue_theme.update(
        background='rgb(28, 31, 48)',
        foreground='rgb(45, 52, 71)',
        primary='rgb(80, 88, 108)',
        current='rgb(184, 112, 0)',
    )

    # Register blue theme
    register_theme('test_blue', blue_theme, 'test')

    # Check that blue theme is listed in available themes
    themes = available_themes()
    assert 'test_blue' in themes

    # Remove theme from available themes
    unregister_theme('test_blue')
    themes = available_themes()
    assert 'test_blue' not in themes


def test_rebuild_theme_settings():
    settings = get_settings()
    assert 'another-theme' not in available_themes()
    # theme is not updated
    with pytest.raises(ValidationError):
        settings.appearance.theme = 'another-theme'
    blue_theme = get_theme('dark').to_rgb_dict()
    register_theme('another-theme', blue_theme, 'test')
    settings.appearance.theme = 'another-theme'


@pytest.mark.parametrize(
    'color',
    [
        '#FF0000',
        'white',
        (0, 127, 127),
        (0, 255, 255, 0.5),
        [50, 200, 200],
        [140, 140, 140, 0.7],
    ],
)
def test_theme(color):
    theme = get_theme('dark')
    theme.background = color


def test_theme_font_size():
    theme = get_theme('dark')
    theme.font_size = '15pt'
    assert theme.font_size == '15pt'

    with pytest.raises(ValidationError):
        theme.font_size = '0pt'
    with pytest.raises(ValidationError):
        theme.font_size = '12px'


def test_theme_syntax_highlight():
    theme = get_theme('dark')
    with pytest.raises(ValidationError):
        theme.syntax_style = 'invalid'


def test_is_theme_available(tmp_path, monkeypatch):
    (tmp_path / 'test_blue').mkdir()
    (tmp_path / 'yellow').mkdir()
    (tmp_path / 'test_blue' / PLUGIN_FILE_NAME).write_text('test-blue')
    monkeypatch.setattr(
        'napari.utils.theme._theme_path', lambda x: tmp_path / x
    )

    n_themes = len(available_themes())

    def mock_install_theme(_themes):
        theme_dict = _themes['dark'].to_rgb_dict()
        theme_dict['id'] = 'test_blue'
        register_theme('test_blue', theme_dict, 'test')

    monkeypatch.setattr(
        'napari.utils.theme._install_npe2_themes', mock_install_theme
    )

    assert len(available_themes()) == n_themes
    assert is_theme_available('dark')
    assert not is_theme_available('green')
    assert not is_theme_available('yellow')
    assert is_theme_available('test_blue')
    assert len(available_themes()) == n_themes + 1
    assert 'test_blue' in available_themes()


@pytest.mark.skipif(
    parse_version(npe2_version) < parse_version('0.6.2'),
    reason='requires npe2 0.6.2 for syntax style contributions',
)
def test_theme_registration(monkeypatch, caplog):
    data = {'dark': get_theme('dark')}

    manifest = PluginManifest(
        name='theme_test',
        display_name='Theme Test',
        contributions=ContributionPoints(
            themes=[
                {
                    'id': 'theme1',
                    'label': 'Theme 1',
                    'type': 'dark',
                    'syntax_style': 'native',
                    'colors': {},
                },
                {
                    'id': 'theme2',
                    'label': 'Theme 2',
                    'type': 'dark',
                    'syntax_style': 'does_not_exist',
                    'colors': {},
                },
            ]
        ),
    )

    def mock_iter_manifests(disabled):
        return [manifest]

    monkeypatch.setattr(
        PluginManager.instance(), 'iter_manifests', mock_iter_manifests
    )
    monkeypatch.setattr('napari.utils.theme._themes', data)
    _install_npe2_themes(data)

    assert 'theme1' in data
    assert 'theme2' not in data
    assert 'Registration theme failed' in caplog.text



"""Image class."""

from __future__ import annotations

import typing
import warnings
from typing import Any, Literal, cast

import numpy as np
from scipy import ndimage as ndi

from napari.layers._data_protocols import LayerDataProtocol
from napari.layers._multiscale_data import MultiScaleData
from napari.layers._scalar_field.scalar_field import ScalarFieldBase
from napari.layers.image._image_constants import (
    ImageProjectionMode,
    ImageRendering,
    Interpolation,
    InterpolationStr,
)
from napari.layers.image._image_utils import guess_rgb
from napari.layers.image._slice import _ImageSliceResponse
from napari.layers.intensity_mixin import IntensityVisualizationMixin
from napari.layers.utils.layer_utils import calc_data_range
from napari.utils._dtype import get_dtype_limits, normalize_dtype
from napari.utils.colormaps import ensure_colormap
from napari.utils.colormaps.colormap_utils import _coerce_contrast_limits
from napari.utils.translations import trans

__all__ = ('Image',)


class Image(IntensityVisualizationMixin, ScalarFieldBase):
    """Image layer.

    Parameters
    ----------
    data : array or list of array
        Image data. Can be N >= 2 dimensional. If the last dimension has length
        3 or 4 can be interpreted as RGB or RGBA if rgb is `True`. If a
        list and arrays are decreasing in shape then the data is treated as
        a multiscale image. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    affine : n-D array or napari.utils.transforms.Affine
        (N+1, N+1) affine transformation matrix in homogeneous coordinates.
        The first (N, N) entries correspond to a linear transform and
        the final column is a length N translation vector and a 1 or a napari
        `Affine` transform object. Applied as an extra transform on top of the
        provided scale, rotate, and shear values.
    attenuation : float
        Attenuation rate for attenuated maximum intensity projection.
    axis_labels : tuple of str
        Dimension names of the layer data.
        If not provided, axis_labels will be set to (..., 'axis -2', 'axis -1').
    blending : str
        One of a list of preset blending modes that determines how RGB and
        alpha values of the layer visual get mixed. Allowed values are
        {'translucent', 'translucent_no_depth', 'additive', 'minimum', 'opaque'}.
    cache : bool
        Whether slices of out-of-core datasets should be cached upon retrieval.
        Currently, this only applies to dask arrays.
    colormap : str, napari.utils.Colormap, tuple, dict
        Colormaps to use for luminance images. If a string, it can be the name
        of a supported colormap from vispy or matplotlib or the name of
        a vispy color or a hexadecimal RGB color representation.
        If a tuple, the first value must be a string to assign as a name to a
        colormap and the second item must be a Colormap. If a dict, the key must
        be a string to assign as a name to a colormap and the value must be a
        Colormap.
    contrast_limits : list (2,)
        Intensity value limits to be used for determining the minimum and maximum colormap bounds for
        luminance images. If not passed, they will be calculated as the min and max intensity value of
        the image.
    custom_interpolation_kernel_2d : np.ndarray
        Convolution kernel used with the 'custom' interpolation mode in 2D rendering.
    depiction : str
        3D Depiction mode. Must be one of {'volume', 'plane'}.
        The default value is 'volume'.
    experimental_clipping_planes : list of dicts, list of ClippingPlane, or ClippingPlaneList
        Each dict defines a clipping plane in 3D in data coordinates.
        Valid dictionary keys are {'position', 'normal', and 'enabled'}.
        Values on the negative side of the normal are discarded if the plane is enabled.
    gamma : float
        Gamma correction for determining colormap linearity; defaults to 1.
    interpolation2d : str
        Interpolation mode used by vispy for rendering 2d data.
        Must be one of our supported modes.
        (for list of supported modes see Interpolation enum)
        'custom' is a special mode for 2D interpolation in which a regular grid
        of samples is taken from the texture around a position using 'linear'
        interpolation before being multiplied with a custom interpolation kernel
        (provided with 'custom_interpolation_kernel_2d').
    interpolation3d : str
        Same as 'interpolation2d' but for 3D rendering.
    iso_threshold : float
        Threshold for isosurface.
    metadata : dict
        Layer metadata.
    multiscale : bool
        Whether the data is a multiscale image or not. Multiscale data is
        represented by a list of array-like image data. If not specified by
        the user and if the data is a list of arrays that decrease in shape,
        then it will be taken to be multiscale. The first image in the list
        should be the largest. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    name : str
        Name of the layer.
    opacity : float
        Opacity of the layer visual, between 0.0 and 1.0.
    plane : dict or SlicingPlane
        Properties defining plane rendering in 3D. Properties are defined in
        data coordinates. Valid dictionary keys are
        {'position', 'normal', 'thickness', and 'enabled'}.
    projection_mode : str
        How data outside the viewed dimensions, but inside the thick Dims slice will
        be projected onto the viewed dimensions. Must fit to ImageProjectionMode
    rendering : str
        Rendering mode used by vispy. Must be one of our supported
        modes.
    rgb : bool, optional
        Whether the image is RGB or RGBA if rgb. If not
        specified by user, but the last dimension of the data has length 3 or 4,
        it will be set as `True`. If `False`, the image is interpreted as a
        luminance image.
    rotate : float, 3-tuple of float, or n-D array.
        If a float, convert into a 2D rotation matrix using that value as an
        angle. If 3-tuple, convert into a 3D rotation matrix, using a yaw,
        pitch, roll convention. Otherwise, assume an nD rotation. Angles are
        assumed to be in degrees. They can be converted from radians with
        'np.degrees' if needed.
    scale : tuple of float
        Scale factors for the layer.
    shear : 1-D array or n-D array
        Either a vector of upper triangular values, or an nD shear matrix with
        ones along the main diagonal.
    translate : tuple of float
        Translation values for the layer.
    units : tuple of str or pint.Unit, optional
        Units of the layer data in world coordinates.
        If not provided, the default units are assumed to be pixels.
    visible : bool
        Whether the layer visual is currently being displayed.

    Attributes
    ----------
    data : array or list of array
        Image data. Can be N dimensional. If the last dimension has length
        3 or 4 can be interpreted as RGB or RGBA if rgb is `True`. If a list
        and arrays are decreasing in shape then the data is treated as a
        multiscale image. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    axis_labels : tuple of str
        Dimension names of the layer data.
    metadata : dict
        Image metadata.
    rgb : bool
        Whether the image is rgb RGB or RGBA if rgb. If not
        specified by user and the last dimension of the data has length 3 or 4
        it will be set as `True`. If `False` the image is interpreted as a
        luminance image.
    multiscale : bool
        Whether the data is a multiscale image or not. Multiscale data is
        represented by a list of array like image data. The first image in the
        list should be the largest. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    mode : str
        Interactive mode. The normal, default mode is PAN_ZOOM, which
        allows for normal interactivity with the canvas.

        In TRANSFORM mode the image can be transformed interactively.
    colormap : 2-tuple of str, napari.utils.Colormap
        The first is the name of the current colormap, and the second value is
        the colormap. Colormaps are used for luminance images, if the image is
        rgb the colormap is ignored.
    colormaps : tuple of str
        Names of the available colormaps.
    contrast_limits : list (2,) of float
        Color limits to be used for determining the colormap bounds for
        luminance images. If the image is rgb the contrast_limits is ignored.
    contrast_limits_range : list (2,) of float
        Range for the color limits for luminance images. If the image is
        rgb the contrast_limits_range is ignored.
    gamma : float
        Gamma correction for determining colormap linearity.
    interpolation2d : str
        Interpolation mode used by vispy. Must be one of our supported modes.
        'custom' is a special mode for 2D interpolation in which a regular grid
        of samples are taken from the texture around a position using 'linear'
        interpolation before being multiplied with a custom interpolation kernel
        (provided with 'custom_interpolation_kernel_2d').
    interpolation3d : str
        Same as 'interpolation2d' but for 3D rendering.
    rendering : str
        Rendering mode used by vispy. Must be one of our supported
        modes.
    depiction : str
        3D Depiction mode used by vispy. Must be one of our supported modes.
    iso_threshold : float
        Threshold for isosurface.
    attenuation : float
        Attenuation rate for attenuated maximum intensity projection.
    plane : SlicingPlane or dict
        Properties defining plane rendering in 3D. Valid dictionary keys are
        {'position', 'normal', 'thickness'}.
    experimental_clipping_planes : ClippingPlaneList
        Clipping planes defined in data coordinates, used to clip the volume.
    custom_interpolation_kernel_2d : np.ndarray
        Convolution kernel used with the 'custom' interpolation mode in 2D rendering.
    units: tuple of pint.Unit
        Units of the layer data in world coordinates.
    Notes
    -----
    _data_view : array (N, M), (N, M, 3), or (N, M, 4)
        Image data for the currently viewed slice. Must be 2D image data, but
        can be multidimensional for RGB or RGBA images if multidimensional is
        `True`.
    """

    _projectionclass = ImageProjectionMode

    def __init__(
        self,
        data,
        *,
        affine=None,
        attenuation=0.05,
        axis_labels=None,
        blending='translucent',
        cache=True,
        colormap='gray',
        contrast_limits=None,
        custom_interpolation_kernel_2d=None,
        depiction='volume',
        experimental_clipping_planes=None,
        gamma=1.0,
        interpolation2d='nearest',
        interpolation3d='linear',
        iso_threshold=None,
        metadata=None,
        multiscale=None,
        name=None,
        opacity=1.0,
        plane=None,
        projection_mode='none',
        rendering='mip',
        rgb=None,
        rotate=None,
        scale=None,
        shear=None,
        translate=None,
        units=None,
        visible=True,
    ):
        # Determine if rgb
        data_shape = data.shape if hasattr(data, 'shape') else data[0].shape
        if rgb and not guess_rgb(data_shape, min_side_len=0):
            raise ValueError(
                trans._(
                    "'rgb' was set to True but data does not have suitable dimensions."
                )
            )
        if rgb is None:
            rgb = guess_rgb(data_shape)

        self.rgb = rgb
        super().__init__(
            data,
            affine=affine,
            axis_labels=axis_labels,
            blending=blending,
            cache=cache,
            custom_interpolation_kernel_2d=custom_interpolation_kernel_2d,
            depiction=depiction,
            experimental_clipping_planes=experimental_clipping_planes,
            metadata=metadata,
            multiscale=multiscale,
            name=name,
            ndim=len(data_shape) - 1 if rgb else len(data_shape),
            opacity=opacity,
            plane=plane,
            projection_mode=projection_mode,
            rendering=rendering,
            rotate=rotate,
            scale=scale,
            shear=shear,
            translate=translate,
            units=units,
            visible=visible,
        )

        self.rgb = rgb
        self._colormap = ensure_colormap(colormap)
        self._gamma = gamma
        self._interpolation2d = Interpolation.NEAREST
        self._interpolation3d = Interpolation.NEAREST
        self.interpolation2d = interpolation2d
        self.interpolation3d = interpolation3d
        self._attenuation = attenuation

        # Set contrast limits, colormaps and plane parameters
        if contrast_limits is None:
            if not isinstance(data, np.ndarray):
                dtype = normalize_dtype(getattr(data, 'dtype', np.float32))
                if np.issubdtype(dtype, np.integer):
                    self.contrast_limits_range = get_dtype_limits(dtype)
                else:
                    self.contrast_limits_range = (0, 1)
                self._should_calc_clims = dtype != np.uint8
            else:
                self.contrast_limits_range = self._calc_data_range()
        else:
            self.contrast_limits_range = contrast_limits
        self._contrast_limits: tuple[float, float] = self.contrast_limits_range
        self.contrast_limits = self._contrast_limits

        if iso_threshold is None:
            cmin, cmax = self.contrast_limits_range
            self._iso_threshold = cmin + (cmax - cmin) / 2
        else:
            self._iso_threshold = iso_threshold

    @property
    def rendering(self):
        """Return current rendering mode.

        Selects a preset rendering mode in vispy that determines how
        volume is displayed.  Options include:

        * ``translucent``: voxel colors are blended along the view ray until
            the result is opaque.
        * ``mip``: maximum intensity projection. Cast a ray and display the
            maximum value that was encountered.
        * ``minip``: minimum intensity projection. Cast a ray and display the
            minimum value that was encountered.
        * ``attenuated_mip``: attenuated maximum intensity projection. Cast a
            ray and attenuate values based on integral of encountered values,
            display the maximum value that was encountered after attenuation.
            This will make nearer objects appear more prominent.
        * ``additive``: voxel colors are added along the view ray until
            the result is saturated.
        * ``iso``: isosurface. Cast a ray until a certain threshold is
            encountered. At that location, lighning calculations are
            performed to give the visual appearance of a surface.
        * ``average``: average intensity projection. Cast a ray and display the
            average of values that were encountered.

        Returns
        -------
        str
            The current rendering mode
        """
        return str(self._rendering)

    @rendering.setter
    def rendering(self, rendering):
        self._rendering = ImageRendering(rendering)
        self.events.rendering()

    def _get_state(self) -> dict[str, Any]:
        """Get dictionary of layer state.

        Returns
        -------
        state : dict of str to Any
            Dictionary of layer state.
        """
        state = self._get_base_state()
        state.update(
            {
                'rgb': self.rgb,
                'multiscale': self.multiscale,
                'colormap': self.colormap.dict(),
                'contrast_limits': self.contrast_limits,
                'interpolation2d': self.interpolation2d,
                'interpolation3d': self.interpolation3d,
                'rendering': self.rendering,
                'depiction': self.depiction,
                'plane': self.plane.dict(),
                'iso_threshold': self.iso_threshold,
                'attenuation': self.attenuation,
                'gamma': self.gamma,
                'data': self.data,
                'custom_interpolation_kernel_2d': self.custom_interpolation_kernel_2d,
            }
        )
        return state

    def _update_slice_response(self, response: _ImageSliceResponse) -> None:
        if self._keep_auto_contrast:
            data = response.image.raw
            input_data = data[-1] if self.multiscale else data
            self.contrast_limits = calc_data_range(
                typing.cast(LayerDataProtocol, input_data), rgb=self.rgb
            )

        super()._update_slice_response(response)

        # Maybe reset the contrast limits based on the new slice.
        if self._should_calc_clims:
            self.reset_contrast_limits_range()
            self.reset_contrast_limits()
            self._should_calc_clims = False
        elif self._keep_auto_contrast:
            self.reset_contrast_limits()

    @property
    def attenuation(self) -> float:
        """float: attenuation rate for attenuated_mip rendering."""
        return self._attenuation

    @attenuation.setter
    def attenuation(self, value: float) -> None:
        self._attenuation = value
        self._update_thumbnail()
        self.events.attenuation()

    @property
    def data(self) -> LayerDataProtocol | MultiScaleData:
        """Data, possibly in multiscale wrapper. Obeys LayerDataProtocol."""
        return self._data

    @data.setter
    def data(self, data: LayerDataProtocol | MultiScaleData) -> None:
        self._data_raw = data
        # note, we don't support changing multiscale in an Image instance
        self._data = MultiScaleData(data) if self.multiscale else data  # type: ignore
        self._update_dims()
        if self._keep_auto_contrast:
            self.reset_contrast_limits()
        self.events.data(value=self.data)
        self._reset_editable()

    @property
    def interpolation2d(self) -> InterpolationStr:
        return cast(InterpolationStr, str(self._interpolation2d))

    @interpolation2d.setter
    def interpolation2d(self, value: InterpolationStr | Interpolation) -> None:
        if value == 'bilinear':
            raise ValueError(
                trans._(
                    "'bilinear' interpolation is not valid for interpolation2d. Did you mean 'linear' instead ?",
                ),
            )
        if value == 'bicubic':
            value = 'cubic'
            warnings.warn(
                trans._("'bicubic' is deprecated. Please use 'cubic' instead"),
                category=DeprecationWarning,
                stacklevel=2,
            )
        self._interpolation2d = Interpolation(value)
        self.events.interpolation2d(value=self._interpolation2d)
        self.events.interpolation(value=self._interpolation2d)

    @property
    def interpolation3d(self) -> InterpolationStr:
        return cast(InterpolationStr, str(self._interpolation3d))

    @interpolation3d.setter
    def interpolation3d(self, value: InterpolationStr | Interpolation) -> None:
        if value == 'custom':
            raise NotImplementedError(
                'custom interpolation is not implemented yet for 3D rendering'
            )
        if value == 'bicubic':
            value = 'cubic'
            warnings.warn(
                trans._("'bicubic' is deprecated. Please use 'cubic' instead"),
                category=DeprecationWarning,
                stacklevel=2,
            )
        self._interpolation3d = Interpolation(value)
        self.events.interpolation3d(value=self._interpolation3d)
        self.events.interpolation(value=self._interpolation3d)

    @property
    def iso_threshold(self) -> float:
        """float: threshold for isosurface."""
        return self._iso_threshold

    @iso_threshold.setter
    def iso_threshold(self, value: float) -> None:
        self._iso_threshold = value
        self._update_thumbnail()
        self.events.iso_threshold()

    def _get_level_shapes(self):
        shapes = super()._get_level_shapes()
        if self.rgb:
            shapes = [s[:-1] for s in shapes]
        return shapes

    def _update_thumbnail(self):
        """Update thumbnail with current image data and colormap."""
        # don't bother updating thumbnail if we don't have any data
        # this also avoids possible dtype mismatch issues below
        # for example np.clip may raise an OverflowError (in numpy 2.0)
        if self._slice.empty:
            return

        image = self._slice.thumbnail.raw

        if self._slice_input.ndisplay == 3 and self.ndim > 2:
            image = np.max(image, axis=0)

        # float16 not supported by ndi.zoom
        dtype = np.dtype(image.dtype)
        if dtype in [np.dtype(np.float16)]:
            image = image.astype(np.float32)

        raw_zoom_factor = np.divide(
            self._thumbnail_shape[:2], image.shape[:2]
        ).min()
        new_shape = np.clip(
            raw_zoom_factor * np.array(image.shape[:2]),
            1,  # smallest side should be 1 pixel wide
            self._thumbnail_shape[:2],
        )
        zoom_factor = tuple(new_shape / image.shape[:2])
        if self.rgb:
            downsampled = ndi.zoom(
                image, zoom_factor + (1,), prefilter=False, order=0
            )
            if image.shape[2] == 4:  # image is RGBA
                colormapped = np.copy(downsampled)
                colormapped[..., 3] = downsampled[..., 3] * self.opacity
                if downsampled.dtype == np.uint8:
                    colormapped = colormapped.astype(np.uint8)
            else:  # image is RGB
                if downsampled.dtype == np.uint8:
                    alpha = np.full(
                        downsampled.shape[:2] + (1,),
                        int(255 * self.opacity),
                        dtype=np.uint8,
                    )
                else:
                    alpha = np.full(downsampled.shape[:2] + (1,), self.opacity)
                colormapped = np.concatenate([downsampled, alpha], axis=2)
        else:
            downsampled = ndi.zoom(
                image, zoom_factor, prefilter=False, order=0
            )
            low, high = self.contrast_limits
            if np.issubdtype(downsampled.dtype, np.integer):
                low = max(low, np.iinfo(downsampled.dtype).min)
                high = min(high, np.iinfo(downsampled.dtype).max)
            downsampled = np.clip(downsampled, low, high)
            color_range = high - low
            if color_range != 0:
                downsampled = (downsampled - low) / color_range
            downsampled = downsampled**self.gamma
            color_array = self.colormap.map(downsampled.ravel())
            colormapped = color_array.reshape((*downsampled.shape, 4))
            colormapped[..., 3] *= self.opacity
        self.thumbnail = colormapped

    def _calc_data_range(
        self, mode: Literal['data', 'slice'] = 'data'
    ) -> tuple[float, float]:
        """
        Calculate the range of the data values in the currently viewed slice
        or full data array
        """
        if mode == 'data':
            input_data = self.data[-1] if self.multiscale else self.data
        elif mode == 'slice':
            data = self._slice.image.raw  # ugh
            input_data = data[-1] if self.multiscale else data
        else:
            raise ValueError(
                trans._(
                    "mode must be either 'data' or 'slice', got {mode!r}",
                    deferred=True,
                    mode=mode,
                )
            )
        return calc_data_range(
            cast(LayerDataProtocol, input_data), rgb=self.rgb
        )

    def _raw_to_displayed(self, raw: np.ndarray) -> np.ndarray:
        """Determine displayed image from raw image.

        This function checks if current contrast_limits are within the range
        supported by vispy.
        If yes, it returns the raw image.
        If not, it rescales the raw image to fit within
        the range supported by vispy.

        Parameters
        ----------
        raw : array
            Raw array.

        Returns
        -------
        image : array
            Displayed array.
        """
        fixed_contrast_info = _coerce_contrast_limits(self.contrast_limits)
        if np.allclose(
            fixed_contrast_info.contrast_limits, self.contrast_limits
        ):
            return raw

        return fixed_contrast_info.coerce_data(raw)

    @IntensityVisualizationMixin.contrast_limits.setter  # type: ignore [attr-defined]
    def contrast_limits(self, contrast_limits):
        IntensityVisualizationMixin.contrast_limits.fset(self, contrast_limits)
        if not np.allclose(
            _coerce_contrast_limits(self.contrast_limits).contrast_limits,
            self.contrast_limits,
        ):
            prev = self._keep_auto_contrast
            self._keep_auto_contrast = False
            try:
                self.refresh(highlight=False, extent=False)
            finally:
                self._keep_auto_contrast = prev

    def _calculate_value_from_ray(self, values):
        # translucent is special: just return the first value, no matter what
        if self.rendering == ImageRendering.TRANSLUCENT:
            return np.ravel(values)[0]
        # iso is weird too: just return None always
        if self.rendering == ImageRendering.ISO:
            return None

        # if the whole ray is NaN, we should see nothing, so return None
        # this check saves us some warnings later as well, so better do it now
        if np.all(np.isnan(values)):
            return None

        # "summary" renderings; they do not represent a specific pixel, so we just
        # return the summary value. We should probably differentiate these somehow.
        # these are also probably not the same as how the gpu does it...
        if self.rendering == ImageRendering.AVERAGE:
            return np.nanmean(values)
        if self.rendering == ImageRendering.ADDITIVE:
            # TODO: this is "broken" cause same pixel gets multisampled...
            #       but it looks like it's also overdoing it in vispy vis too?
            #       I don't know if there's a way to *not* do it...
            return np.nansum(values)

        # all the following cases are returning the *actual* value of the image at the
        # "selected" pixel, whose position changes depending on the rendering mode.
        if self.rendering == ImageRendering.MIP:
            return np.nanmax(values)
        if self.rendering == ImageRendering.MINIP:
            return np.nanmin(values)
        if self.rendering == ImageRendering.ATTENUATED_MIP:
            # normalize values so attenuation applies from 0 to 1
            values_attenuated = (
                values - self.contrast_limits[0]
            ) / self.contrast_limits[1]
            # approx, step size is actually calculated with int(lenght(ray) * 2)
            step_size = 0.5
            sumval = (
                step_size
                * np.cumsum(np.clip(values_attenuated, 0, 1))
                * len(values_attenuated)
            )
            scale = np.exp(-self.attenuation * (sumval - 1))
            return values[np.nanargmin(values_attenuated * scale)]

        raise RuntimeError(  # pragma: no cover
            f'ray value calculation not implemented for {self.rendering}'
        )



from napari._qt.containers._factory import create_model, create_view
from napari._qt.containers.qt_axis_model import (
    AxisList,
    AxisModel,
    QtAxisListModel,
)
from napari._qt.containers.qt_layer_list import QtLayerList
from napari._qt.containers.qt_layer_model import QtLayerListModel
from napari._qt.containers.qt_list_model import QtListModel
from napari._qt.containers.qt_list_view import QtListView
from napari._qt.containers.qt_tree_model import QtNodeTreeModel
from napari._qt.containers.qt_tree_view import QtNodeTreeView

__all__ = [
    'AxisList',
    'AxisModel',
    'QtAxisListModel',
    'QtLayerList',
    'QtLayerListModel',
    'QtListModel',
    'QtListView',
    'QtNodeTreeModel',
    'QtNodeTreeView',
    'create_model',
    'create_view',
]



"""
This module test some of the behavior of action manager.
"""

from unittest.mock import Mock

import pytest

from napari.utils.action_manager import ActionManager


@pytest.fixture
def action_manager():
    """
    Unlike normal napari we use a different instance we have complete control
    over and can throw away this makes it easier.
    """
    return ActionManager()


def test_unbind_non_existing_action(action_manager):
    """
    We test that unbinding an non existing action is ok, this can happen due to
    keybindings in settings while some plugins are deactivated or upgraded.

    We emit a warning but should not fail.
    """
    with pytest.warns(UserWarning):
        assert action_manager.unbind_shortcut('napari:foo_bar') is None


def test_bind_multiple_action(action_manager):
    """
    Test we can have multiple bindings per action
    """

    action_manager.register_action(
        'napari:test_action_2', lambda: None, 'this is a test action', None
    )

    action_manager.bind_shortcut('napari:test_action_2', 'X')
    action_manager.bind_shortcut('napari:test_action_2', 'Y')
    assert action_manager._shortcuts['napari:test_action_2'] == ['X', 'Y']


def test_bind_unbind_existing_action(action_manager):
    action_manager.register_action(
        'napari:test_action_1', lambda: None, 'this is a test action', None
    )

    assert action_manager.bind_shortcut('napari:test_action_1', 'X') is None
    assert action_manager.unbind_shortcut('napari:test_action_1') == ['X']
    assert action_manager._shortcuts['napari:test_action_1'] == []


def test_bind_key_generator(action_manager):
    def _sample_generator():
        yield 'X'

    action_manager.register_action(
        'napari:test_action_1',
        _sample_generator,
        'this is a test action',
        None,
    )

    with pytest.raises(ValueError, match='generator functions'):
        action_manager.bind_button('napari:test_action_1', Mock())



import numpy as np

from napari.settings._application import (
    GridHeight,
    GridSpacing,
    GridStride,
    GridWidth,
)
from napari.utils.events import EventedModel


class GridCanvas(EventedModel):
    """Grid for canvas.

    Right now the only grid mode that is still inside one canvas with one
    camera, but future grid modes could support multiple canvases.

    Attributes
    ----------
    enabled : bool
        If grid is enabled or not.
    stride : int
        Number of layers to place in each grid square before moving on to
        the next square. The default ordering is to place the most visible
        layer in the top left corner of the grid. A negative stride will
        cause the order in which the layers are placed in the grid to be
        reversed.
    shape : 2-tuple of int
        Number of rows and columns in the grid. A value of -1 for either or
        both of will be used the row and column numbers will trigger an
        auto calculation of the necessary grid shape to appropriately fill
        all the layers at the appropriate stride.
    spacing : float
        Spacing between grid layers, as a proportion of the average
        of the height and width of the extent of layers in world coordinates after slicing.
        A value of 0.0 will have the grid layers touching each other.
        Positive values will space the layers apart, and negative values
        will overlap the layers.

        .. versionadded:: 0.6.0
            ``spacing`` was added in 0.6.0.
    """

    # fields
    # See https://github.com/pydantic/pydantic/issues/156 for why
    # these need a type: ignore comment
    stride: GridStride = 1  # type: ignore[valid-type]
    shape: tuple[GridHeight, GridWidth] = (-1, -1)  # type: ignore[valid-type]
    enabled: bool = False
    spacing: GridSpacing = 0.0  # type: ignore[valid-type]

    def actual_shape(self, nlayers: int = 1) -> tuple[int, int]:
        """Return the actual shape of the grid.

        This will return the shape parameter, unless one of the row
        or column numbers is -1 in which case it will compute the
        optimal shape of the grid given the number of layers and
        current stride.

        If the grid is not enabled, this will return (1, 1).

        Parameters
        ----------
        nlayers : int
            Number of layers that need to be placed in the grid.

        Returns
        -------
        shape : 2-tuple of int
            Number of rows and columns in the grid.
        """
        if not self.enabled:
            return (1, 1)

        if nlayers == 0:
            return (1, 1)

        n_row, n_column = self.shape
        n_grid_squares = np.ceil(nlayers / abs(self.stride)).astype(int)

        if n_row == -1 and n_column == -1:
            n_column = np.ceil(np.sqrt(n_grid_squares)).astype(int)
            n_row = np.ceil(n_grid_squares / n_column).astype(int)
        elif n_row == -1:
            n_row = np.ceil(n_grid_squares / n_column).astype(int)
        elif n_column == -1:
            n_column = np.ceil(n_grid_squares / n_row).astype(int)

        n_row = max(1, n_row)
        n_column = max(1, n_column)

        return (n_row, n_column)

    def position(self, index: int, nlayers: int) -> tuple[int, int]:
        """Return the position of a given linear index in grid.

        If the grid is not enabled, this will return (0, 0).

        Parameters
        ----------
        index : int
            Position of current layer in layer list.
        nlayers : int
            Number of layers that need to be placed in the grid.

        Returns
        -------
        position : 2-tuple of int
            Row and column position of current index in the grid.
        """
        if not self.enabled:
            return (0, 0)

        n_row, n_column = self.actual_shape(nlayers)

        # Adjust for forward or reverse ordering
        adj_i = nlayers - index - 1 if self.stride < 0 else index

        adj_i = adj_i // abs(self.stride)
        adj_i = adj_i % (n_row * n_column)
        i_row = adj_i // n_column
        i_column = adj_i % n_column
        return (i_row, i_column)



import pytest

from napari._app_model import get_app_model
from napari._qt._qapp_model.qactions._window import toggle_action_details
from napari._tests.utils import skip_local_popups


@skip_local_popups
@pytest.mark.parametrize(
    (
        'action_id',
        'action_text',
        'action_dockwidget_name',
        'action_status_tooltip',
    ),
    toggle_action_details,
)
def test_toggle_dockwidget_actions(
    make_napari_viewer,
    action_id,
    action_text,
    action_dockwidget_name,
    action_status_tooltip,
):
    app = get_app_model()
    viewer = make_napari_viewer(show=True)
    widget = getattr(viewer.window._qt_viewer, action_dockwidget_name)
    widget_initial_visibility = widget.isVisible()
    action = viewer.window.window_menu.findAction(action_id)

    # ---- Check initial action checked state
    # Need to emit manually menu `aboutToShow` signal to ensure actions checked state is synced
    viewer.window.window_menu.aboutToShow.emit()
    # If the action is checked the widget should be visible
    # If the action is not checked the widget shouldn't be visible
    assert action.isChecked() == widget.isVisible()

    # ---- Check toggling dockwidget from initial visibility
    app.commands.execute_command(action_id)
    assert widget.isVisible() != widget_initial_visibility
    viewer.window.window_menu.aboutToShow.emit()
    assert action.isChecked() == widget.isVisible()

    # ---- Check restoring initial visibility
    app.commands.execute_command(action_id)
    assert widget.isVisible() == widget_initial_visibility
    viewer.window.window_menu.aboutToShow.emit()
    assert action.isChecked() == widget.isVisible()



"""
This module provides compatibility between pydantic v1 and v2.

Keep using this compatibility module until we stop using any pydantic v1 API functionality.
This can be removed when everything has been migrated to pydantic v2.
"""

# The Pydantic V2 package can access the Pydantic V1 API by importing through `pydantic.v1`.
# See https://docs.pydantic.dev/latest/migration/#continue-using-pydantic-v1-features
from pydantic.v1 import (
    BaseModel,
    BaseSettings,
    Extra,
    Field,
    PrivateAttr,
    ValidationError,
    color,
    conlist,
    constr,
    errors,
    main,
    parse_obj_as,
    root_validator,
    types,
    utils,
    validator,
)
from pydantic.v1.env_settings import (
    EnvSettingsSource,
    SettingsError,
    SettingsSourceCallable,
)
from pydantic.v1.error_wrappers import ErrorWrapper, display_errors
from pydantic.v1.fields import SHAPE_LIST, ModelField
from pydantic.v1.generics import GenericModel
from pydantic.v1.main import ClassAttribute, ModelMetaclass
from pydantic.v1.utils import ROOT_KEY, sequence_like

Color = color.Color

__all__ = (
    'ROOT_KEY',
    'SHAPE_LIST',
    'BaseModel',
    'BaseSettings',
    'ClassAttribute',
    'Color',
    'EnvSettingsSource',
    'ErrorWrapper',
    'Extra',
    'Field',
    'GenericModel',
    'ModelField',
    'ModelMetaclass',
    'PrivateAttr',
    'SettingsError',
    'SettingsSourceCallable',
    'ValidationError',
    'color',
    'conlist',
    'constr',
    'display_errors',
    'errors',
    'main',
    'parse_obj_as',
    'root_validator',
    'sequence_like',
    'types',
    'utils',
    'validator',
)



"""Miscellaneous utility functions."""

from __future__ import annotations

import builtins
import collections.abc
import contextlib
import inspect
import itertools
import os
import re
import sys
import warnings
from collections.abc import Callable, Iterable, Iterator, Sequence
from enum import Enum, EnumMeta
from os import fspath, path as os_path
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

import numpy as np
import numpy.typing as npt

from napari.utils.translations import trans

_sentinel = object()

if TYPE_CHECKING:
    import packaging.version


ROOT_DIR = os_path.dirname(os_path.dirname(__file__))


def parse_version(v: str) -> packaging.version._BaseVersion:
    """Parse a version string and return a packaging.version.Version obj."""
    import packaging.version

    try:
        return packaging.version.Version(v)
    except packaging.version.InvalidVersion:
        return packaging.version.LegacyVersion(v)  # type: ignore[attr-defined]


def running_as_constructor_app() -> bool:
    """Infer whether we are running as a constructor bundle."""
    return (
        Path(sys.prefix).parent.parent / '.napari_is_bundled_constructor'
    ).exists()


def in_jupyter() -> bool:
    """Return true if we're running in jupyter notebook/lab or qtconsole."""
    with contextlib.suppress(ImportError):
        from IPython import get_ipython

        return get_ipython().__class__.__name__ == 'ZMQInteractiveShell'
    return False


def in_ipython() -> bool:
    """Return true if we're running in an IPython interactive shell."""
    with contextlib.suppress(ImportError):
        from IPython import get_ipython

        return get_ipython().__class__.__name__ == 'TerminalInteractiveShell'
    return False


def in_python_repl() -> bool:
    """Return true if we're running in a Python REPL."""
    with contextlib.suppress(ImportError):
        from IPython import get_ipython

        return get_ipython().__class__.__name__ == 'NoneType' and hasattr(
            sys, 'ps1'
        )
    return False


def str_to_rgb(arg: str) -> list[int]:
    """Convert an rgb string 'rgb(x,y,z)' to a list of ints [x,y,z]."""
    match = re.match(r'rgb\((\d+),\s*(\d+),\s*(\d+)\)', arg)
    if match is None:
        raise ValueError("arg not in format 'rgb(x,y,z)'")
    return list(map(int, match.groups()))


def ensure_iterable(
    arg: None | str | Enum | float | list | npt.NDArray,
):
    """Ensure an argument is an iterable. Useful when an input argument
    can either be a single value or a list.
    """
    if is_iterable(arg):
        return arg

    return itertools.repeat(arg)


def is_iterable(
    arg: None | str | Enum | float | list | npt.NDArray,
    allow_none: bool = False,
) -> bool:
    """Determine if a single argument is an iterable."""
    if arg is None:
        return allow_none

    # Here if arg is None it used to return allow_none
    if isinstance(arg, str | Enum) or np.isscalar(arg):
        return False

    return isinstance(arg, collections.abc.Iterable)


def is_sequence(arg: Any) -> bool:
    """Check if ``arg`` is a sequence like a list or tuple.

    return True:
        list
        tuple
    return False:
        string
        numbers
        dict
        set
    """
    return bool(
        isinstance(arg, collections.abc.Sequence) and not isinstance(arg, str)
    )


def ensure_sequence_of_iterables(
    obj: Any,
    length: int | None = None,
    repeat_empty: bool = False,
    allow_none: bool = False,
):
    """Ensure that ``obj`` behaves like a (nested) sequence of iterables.

    If length is provided and the object is already a sequence of iterables,
    a ValueError will be raised if ``len(obj) != length``.

    Parameters
    ----------
    obj : Any
        the object to check
    length : int, optional
        If provided, assert that obj has len ``length``, by default None
    repeat_empty : bool
        whether to repeat an empty sequence (otherwise return the empty sequence itself)
    allow_none : bool
        treat None as iterable

    Returns
    -------
    iterable
        nested sequence of iterables, or an itertools.repeat instance

    Examples
    --------
    In [1]: ensure_sequence_of_iterables([1, 2])
    Out[1]: repeat([1, 2])

    In [2]: ensure_sequence_of_iterables([(1, 2), (3, 4)])
    Out[2]: [(1, 2), (3, 4)]

    In [3]: ensure_sequence_of_iterables([(1, 2), None], allow_none=True)
    Out[3]: [(1, 2), None]

    In [4]: ensure_sequence_of_iterables({'a':1})
    Out[4]: repeat({'a': 1})

    In [5]: ensure_sequence_of_iterables(None)
    Out[5]: repeat(None)

    In [6]: ensure_sequence_of_iterables([])
    Out[6]: repeat([])

    In [7]: ensure_sequence_of_iterables([], repeat_empty=False)
    Out[7]: []
    """

    if (
        obj is not None
        and is_sequence(obj)
        and all(is_iterable(el, allow_none=allow_none) for el in obj)
        and (not repeat_empty or len(obj) > 0)
    ):
        if length is not None and len(obj) != length:
            # sequence of iterables of wrong length
            raise ValueError(
                trans._(
                    'length of {obj} must equal {length}',
                    deferred=True,
                    obj=obj,
                    length=length,
                )
            )

        if len(obj) > 0 or not repeat_empty:
            return obj

    return itertools.repeat(obj)


def formatdoc(obj):
    """Substitute globals and locals into an object's docstring."""
    frame = inspect.currentframe().f_back
    try:
        obj.__doc__ = obj.__doc__.format(
            **{**frame.f_globals, **frame.f_locals}
        )
    finally:
        del frame
    return obj


class StringEnumMeta(EnumMeta):
    def __getitem__(self, item):
        """set the item name case to uppercase for name lookup"""
        if isinstance(item, str):
            item = item.upper()

        return super().__getitem__(item)

    def __call__(
        cls,
        value,
        names=None,
        *,
        module=None,
        qualname=None,
        type=None,  # noqa: A002
        start=1,
    ):
        """set the item value case to lowercase for value lookup"""
        # simple value lookup
        if names is None:
            if isinstance(value, str):
                return super().__call__(value.lower())
            if isinstance(value, cls):
                return value

            raise ValueError(
                trans._(
                    '{class_name} may only be called with a `str` or an instance of {class_name}. Got {dtype}',
                    deferred=True,
                    class_name=cls,
                    dtype=builtins.type(value),
                )
            )

        # otherwise create new Enum class
        return cls._create_(
            value,
            names,
            module=module,
            qualname=qualname,
            type=type,
            start=start,
        )

    def keys(self) -> list[str]:
        return list(map(str, self))


class StringEnum(Enum, metaclass=StringEnumMeta):
    @staticmethod
    def _generate_next_value_(name: str, start, count, last_values) -> str:
        """autonaming function assigns each value its own name as a value"""
        return name.lower()

    def __str__(self) -> str:
        """String representation: The string method returns the lowercase
        string of the Enum name
        """
        return self.value

    def __eq__(self, other: object) -> bool:
        if type(self) is type(other):
            return self is other
        if isinstance(other, str):
            return str(self) == other
        return False

    def __hash__(self) -> int:
        return hash(str(self))


camel_to_snake_pattern = re.compile(r'(.)([A-Z][a-z]+)')
camel_to_spaces_pattern = re.compile(
    r'((?<=[a-z])[A-Z]|(?<!\A)[A-R,T-Z](?=[a-z]))'
)


def camel_to_snake(name: str) -> str:
    # https://gist.github.com/jaytaylor/3660565
    return camel_to_snake_pattern.sub(r'\1_\2', name).lower()


def camel_to_spaces(val: str) -> str:
    return camel_to_spaces_pattern.sub(r' \1', val)


T = TypeVar('T', str, Path)


def abspath_or_url(relpath: T, *, must_exist: bool = False) -> T:
    """Utility function that normalizes paths or a sequence thereof.

    Expands user directory and converts relpaths to abspaths... but ignores
    URLS that begin with "http", "ftp", or "file".

    Parameters
    ----------
    relpath : str|Path
        A path, either as string or Path object.
    must_exist : bool, default True
        Raise ValueError if `relpath` is not a URL and does not exist.

    Returns
    -------
    abspath : str|Path
        An absolute path, or list or tuple of absolute paths (same type as
        input)
    """
    from urllib.parse import urlparse

    if not isinstance(relpath, str | Path):
        raise TypeError(
            trans._('Argument must be a string or Path', deferred=True)
        )
    OriginType = type(relpath)

    relpath_str = fspath(relpath)
    urlp = urlparse(relpath_str)
    if urlp.scheme and urlp.netloc:
        return OriginType(relpath_str)

    path = os_path.abspath(os_path.expanduser(relpath_str))
    if must_exist and not (urlp.scheme or urlp.netloc or os.path.exists(path)):
        raise ValueError(
            trans._(
                'Requested path {path!r} does not exist.',
                deferred=True,
                path=path,
            )
        )
    return OriginType(path)


def all_subclasses(cls: type) -> set:
    """Recursively find all subclasses of class ``cls``.

    Parameters
    ----------
    cls : class
        A python class (or anything that implements a __subclasses__ method).

    Returns
    -------
    set
        the set of all classes that are subclassed from ``cls``
    """
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


def ensure_n_tuple(val: Iterable, n: int, fill: int = 0) -> tuple:
    """Ensure input is a length n tuple.

    Parameters
    ----------
    val : iterable
        Iterable to be forced into length n-tuple.
    n : int
        Length of tuple.

    Returns
    -------
    tuple
        Coerced tuple.
    """
    assert n > 0, 'n must be greater than 0'
    tuple_value = tuple(val)
    return (fill,) * (n - len(tuple_value)) + tuple_value[-n:]


def ensure_layer_data_tuple(val: tuple) -> tuple:
    msg = trans._(
        'Not a valid layer data tuple: {value!r}',
        deferred=True,
        value=val,
    )
    if not isinstance(val, tuple) and val:
        raise TypeError(msg)
    if len(val) > 1:
        if not isinstance(val[1], collections.abc.Mapping):
            raise TypeError(msg)
        if len(val) > 2 and not isinstance(val[2], str):
            raise TypeError(msg)
    return val


def ensure_list_of_layer_data_tuple(val: list[tuple]) -> list[tuple]:
    # allow empty list to be returned but do nothing in that case
    if isinstance(val, list):
        with contextlib.suppress(TypeError):
            return [ensure_layer_data_tuple(v) for v in val]
    raise TypeError(
        trans._('Not a valid list of layer data tuples!', deferred=True)
    )


def _quiet_array_equal(*a, **k) -> bool:
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'elementwise comparison')
        return np.array_equal(*a, **k)


def _pandas_dataframe_equal(df1, df2):
    return df1.equals(df2)


def _arraylike_short_names(obj) -> Iterator[str]:
    """Yield all the short names of an array-like or its class."""
    type_ = type(obj) if not inspect.isclass(obj) else obj
    for base in type_.mro():
        yield f'{base.__module__.split(".", maxsplit=1)[0]}.{base.__name__}'


def pick_equality_operator(obj: Any) -> Callable[[Any, Any], bool]:
    """Return a function that can check equality between ``obj`` and another.

    Rather than always using ``==`` (i.e. ``operator.eq``), this function
    returns operators that are aware of object types: mostly "array types with
    more than one element" whose truth value is ambiguous.

    This function works for both classes (types) and instances.  If an instance
    is passed, it will be first cast to a type with type(obj).

    Parameters
    ----------
    obj : Any
        An object whose equality with another object you want to check.

    Returns
    -------
    operator : Callable[[Any, Any], bool]
        An operation that can be called as ``operator(obj, other)`` to check
        equality between objects of type ``type(obj)``.
    """
    import operator

    # yes, it's a little riskier, but we are checking namespaces instead of
    # actual `issubclass` here to avoid slow import times
    _known_arrays: dict[str, Callable[[Any, Any], bool]] = {
        'numpy.ndarray': _quiet_array_equal,  # numpy.ndarray
        'dask.Array': operator.is_,  # dask.array.core.Array
        'dask.Delayed': operator.is_,  # dask.delayed.Delayed
        'zarr.Array': operator.is_,  # zarr.core.Array
        'xarray.DataArray': _quiet_array_equal,  # xarray.core.dataarray.DataArray
        'pandas.DataFrame': _pandas_dataframe_equal,  # pandas.DataFrame.equals
    }

    for name in _arraylike_short_names(obj):
        func = _known_arrays.get(name)
        if func:
            return func

    return operator.eq


def _is_array_type(array: npt.ArrayLike, type_name: str) -> bool:
    """Checks if an array-like instance or class is of the type described by a short name.

    This is useful when you want to check the type of array-like quickly without
    importing its package, which might take a long time.

    Parameters
    ----------
    array
        The array-like object.
    type_name : str
        The short name of the type to test against (e.g. 'numpy.ndarray', 'xarray.DataArray').

    Returns
    -------
    True if the array is associated with the type name.
    """
    return type_name in _arraylike_short_names(array)


def dir_hash(
    path: str | Path,
    include_paths: bool = True,
    ignore_hidden: bool = True,
) -> str:
    """Compute the hash of a directory, based on structure and contents.

    Parameters
    ----------
    path : Union[str, Path]
        Source path which will be used to select all files (and files in subdirectories)
        to compute the hexadecimal digest.
    include_paths : bool
        If ``True``, the hash will also include the ``file`` parts.
    ignore_hidden : bool
        If ``True``, hidden files (starting with ``.``) will be ignored when
        computing the hash.

    Returns
    -------
    hash : str
        Hexadecimal digest of all files in the provided path.
    """
    import hashlib

    if not Path(path).is_dir():
        raise TypeError(
            trans._(
                '{path} is not a directory.',
                deferred=True,
                path=path,
            )
        )

    hash_func = hashlib.md5
    _hash = hash_func()
    for root, _, files in os.walk(path):
        for fname in sorted(files):
            if fname.startswith('.') and ignore_hidden:
                continue
            _file_hash(_hash, Path(root) / fname, Path(path), include_paths)
    return _hash.hexdigest()


def paths_hash(
    paths: Iterable[str | Path],
    include_paths: bool = True,
    ignore_hidden: bool = True,
) -> str:
    """Compute the hash of list of paths.

    Parameters
    ----------
    paths : Iterable[Union[str, Path]]
        An iterable of paths to files which will be used when computing the hash.
    include_paths : bool
        If ``True``, the hash will also include the ``file`` parts.
    ignore_hidden : bool
        If ``True``, hidden files (starting with ``.``) will be ignored when
        computing the hash.

    Returns
    -------
    hash : str
        Hexadecimal digest of the contents of provided files.
    """
    import hashlib

    hash_func = hashlib.md5
    _hash = hash_func()
    for file_path in sorted(paths):
        file_path = Path(file_path)
        if ignore_hidden and str(file_path.stem).startswith('.'):
            continue
        _file_hash(_hash, file_path, file_path.parent, include_paths)
    return _hash.hexdigest()


def _file_hash(
    _hash, file: Path, path: Path, include_paths: bool = True
) -> None:
    """Update hash with based on file contents and optionally relative path.

    Parameters
    ----------
    _hash
    file : Path
        Path to the source file which will be used to compute the hash.
    path : Path
        Path to the base directory of the `file`. This can be usually obtained by using `file.parent`.
    include_paths : bool
        If ``True``, the hash will also include the ``file`` parts.
    """
    _hash.update(file.read_bytes())

    if include_paths:
        # update the hash with the filename
        fparts = file.relative_to(path).parts
        _hash.update(''.join(fparts).encode())


def _combine_signatures(
    *objects: Callable,
    return_annotation=inspect.Signature.empty,
    exclude: Iterable[str] = (),
) -> inspect.Signature:
    """Create combined Signature from objects, excluding names in `exclude`.

    Parameters
    ----------
    *objects : Callable
        callables whose signatures should be combined
    return_annotation : [type], optional
        The return annotation to use for combined signature, by default
        inspect.Signature.empty (as it's ambiguous)
    exclude : tuple, optional
        Parameter names to exclude from the combined signature (such as
        'self'), by default ()

    Returns
    -------
    inspect.Signature
        Signature object with the combined signature. Reminder, str(signature)
        provides a very nice repr for code generation.
    """
    params = itertools.chain(
        *(inspect.signature(o).parameters.values() for o in objects)
    )
    new_params = sorted(
        (p for p in params if p.name not in exclude),
        key=lambda p: p.kind,
    )
    return inspect.Signature(new_params, return_annotation=return_annotation)


def deep_update(dct: dict, merge_dct: dict, copy: bool = True) -> dict:
    """Merge possibly nested dicts"""
    _dct = dct.copy() if copy else dct
    for k, v in merge_dct.items():
        if k in _dct and isinstance(dct[k], dict) and isinstance(v, dict):
            deep_update(_dct[k], v, copy=False)
        else:
            _dct[k] = v
    return _dct


def install_certifi_opener() -> None:
    """Install urlopener that uses certifi context.

    This is useful in the bundle, where otherwise users might get SSL errors
    when using `urllib.request.urlopen`.
    """
    import ssl
    from urllib import request

    import certifi

    context = ssl.create_default_context(cafile=certifi.where())
    https_handler = request.HTTPSHandler(context=context)
    opener = request.build_opener(https_handler)
    request.install_opener(opener)


def maybe_patch_conda_exe() -> None:
    if sys.platform != 'win32':
        # Patch required only on Windows
        return
    if 'CONDA_EXE' in os.environ and Path(os.environ['CONDA_EXE']).is_file():
        # conde exe already pointed by variable, no need to patch
        return
    if 'CONDA_PYTHON_EXE' not in os.environ:
        # We do not have base path, so nothing to patch
        return
    conda_path = (
        Path(os.getenv('CONDA_PYTHON_EXE', '')).parent
        / 'Scripts'
        / 'conda.exe'
    )
    if not conda_path.is_file():
        # conda.exe does not exist under expected location, so nothing to patch
        return
    os.environ['CONDA_EXE'] = str(conda_path)


def reorder_after_dim_reduction(order: Sequence[int]) -> tuple[int, ...]:
    """Ensure current dimension order is preserved after dims are dropped.

    This is similar to :func:`scipy.stats.rankdata`, but only deals with
    unique integers (like dimension indices), so is simpler and faster.

    Parameters
    ----------
    order : Sequence[int]
        The data to reorder.

    Returns
    -------
    Tuple[int, ...]
        A permutation of ``range(len(order))`` that is consistent with the input order.

    Examples
    --------
    >>> reorder_after_dim_reduction([2, 0])
    (1, 0)

    >>> reorder_after_dim_reduction([0, 1, 2])
    (0, 1, 2)

    >>> reorder_after_dim_reduction([4, 0, 2])
    (2, 0, 1)
    """
    # A single argsort works for strictly increasing/decreasing orders,
    # but not for arbitrary orders.
    return tuple(argsort(argsort(order)))


def argsort(values: Sequence[int]) -> list[int]:
    """Equivalent to :func:`numpy.argsort` but faster in some cases.

    Parameters
    ----------
    values : Sequence[int]
        The integer values to sort.

    Returns
    -------
    List[int]
        The indices that when used to index the input values will produce
        the values sorted in increasing order.

    Examples
    --------
    >>> argsort([2, 0])
    [1, 0]

    >>> argsort([0, 1, 2])
    [0, 1, 2]

    >>> argsort([4, 0, 2])
    [1, 2, 0]
    """
    return sorted(range(len(values)), key=values.__getitem__)



import os
from unittest.mock import Mock

import numpy as np
import pytest

from napari import Viewer, layers
from napari._pydantic_compat import ValidationError
from napari._tests.utils import (
    add_layer_by_type,
    check_view_transform_consistency,
    check_viewer_functioning,
    layer_test_data,
    skip_local_popups,
    skip_on_win_ci,
)
from napari.settings import get_settings
from napari.utils._tests.test_naming import eval_with_filename
from napari.utils.action_manager import action_manager


def _get_provider_actions(type_):
    actions = set()
    for superclass in type_.mro():
        actions.update(
            action.command
            for action in action_manager._get_provider_actions(
                superclass
            ).values()
        )
    return actions


def _assert_shortcuts_exist_for_each_action(type_):
    actions = _get_provider_actions(type_)
    shortcuts = {
        name.partition(':')[-1] for name in get_settings().shortcuts.shortcuts
    }
    shortcuts.update(func.__name__ for func in type_.class_keymap.values())
    for action in actions:
        assert action.__name__ in shortcuts, (
            f"missing shortcut for action '{action.__name__}' on '{type_.__name__}' is missing"
        )


viewer_actions = _get_provider_actions(Viewer)


def test_all_viewer_actions_are_accessible_via_shortcut(make_napari_viewer):
    """
    Make sure we do find all the actions attached to a viewer via keybindings
    """
    # instantiate to make sure everything is initialized correctly
    _ = make_napari_viewer()
    _assert_shortcuts_exist_for_each_action(Viewer)


@pytest.mark.xfail
def test_non_existing_bindings():
    """
    Those are condition tested in next unittest; but do not exists; this is
    likely due to an oversight somewhere.
    """
    assert 'play' in [func.__name__ for func in viewer_actions]
    assert 'toggle_fullscreen' in [func.__name__ for func in viewer_actions]


@pytest.mark.parametrize('func', viewer_actions)
def test_viewer_actions(make_napari_viewer, func):
    viewer = make_napari_viewer()

    if func.__name__ == 'toggle_fullscreen' and not os.getenv('CI'):
        pytest.skip('Fullscreen cannot be tested in CI')
    if func.__name__ == 'play':
        pytest.skip('Play cannot be tested with Pytest')
    func(viewer)


def test_viewer(make_napari_viewer):
    """Test instantiating viewer."""
    viewer = make_napari_viewer()
    view = viewer.window._qt_viewer

    assert viewer.title == 'napari'
    assert view.viewer == viewer

    assert len(viewer.layers) == 0
    assert view.layers.model().rowCount() == 0

    assert viewer.dims.ndim == 2
    assert view.dims.nsliders == viewer.dims.ndim
    assert np.sum(view.dims._displayed_sliders) == 0

    # Switch to 3D rendering mode and back to 2D rendering mode
    viewer.dims.ndisplay = 3
    assert viewer.dims.ndisplay == 3
    viewer.dims.ndisplay = 2
    assert viewer.dims.ndisplay == 2


@pytest.mark.parametrize(('layer_class', 'data', 'ndim'), layer_test_data)
def test_add_layer(make_napari_viewer, layer_class, data, ndim):
    viewer = make_napari_viewer()
    layer = add_layer_by_type(viewer, layer_class, data, visible=True)
    check_viewer_functioning(viewer, viewer.window._qt_viewer, data, ndim)

    for func in layer.class_keymap.values():
        func(layer)


layer_types = (
    'Image',
    'Vectors',
    'Surface',
    'Tracks',
    'Points',
    'Labels',
    'Shapes',
)


@pytest.mark.parametrize(('layer_class', 'data', 'ndim'), layer_test_data)
def test_all_layer_actions_are_accessible_via_shortcut(
    layer_class, data, ndim
):
    """
    Make sure we do find all the actions attached to a layer via keybindings
    """
    # instantiate to make sure everything is initialized correctly
    _ = layer_class(data)
    _assert_shortcuts_exist_for_each_action(layer_class)


@pytest.mark.parametrize(
    ('layer_class', 'a_unique_name', 'ndim'), layer_test_data
)
def test_add_layer_magic_name(
    make_napari_viewer, layer_class, a_unique_name, ndim
):
    """Test magic_name works when using add_* for layers"""
    # Tests for issue #1709
    viewer = make_napari_viewer()  # noqa: F841
    layer = eval_with_filename(
        'add_layer_by_type(viewer, layer_class, a_unique_name)',
        'somefile.py',
    )
    assert layer.name == 'a_unique_name'


@skip_on_win_ci
def test_screenshot(make_napari_viewer, qtbot):
    """Test taking a screenshot."""
    viewer = make_napari_viewer()

    np.random.seed(0)
    # Add image
    data = np.random.random((10, 15))
    viewer.add_image(data)

    # Add labels
    data = np.random.randint(20, size=(10, 15))
    viewer.add_labels(data)

    # Add points
    data = 20 * np.random.random((10, 2))
    viewer.add_points(data)

    # Add vectors
    data = 20 * np.random.random((10, 2, 2))
    viewer.add_vectors(data)

    # Add shapes
    data = 20 * np.random.random((10, 4, 2))
    viewer.add_shapes(data)

    # Take screenshot of the image canvas only
    # Test that flash animation does not occur
    screenshot = viewer.screenshot(canvas_only=True)
    assert not hasattr(
        viewer.window._qt_viewer._welcome_widget, '_flash_animation'
    )
    assert screenshot.ndim == 3

    # Take screenshot with the viewer included
    screenshot = viewer.screenshot(canvas_only=False)
    assert screenshot.ndim == 3

    # test size argument (and ensure it coerces to int)
    screenshot = viewer.screenshot(canvas_only=True, size=(20, 20.0))
    assert screenshot.shape == (20, 20, 4)

    # test flash animation works
    screenshot = viewer.screenshot(canvas_only=True, flash=True)
    assert hasattr(
        viewer.window._qt_viewer._welcome_widget, '_flash_animation'
    )

    # Here we wait until the flash animation will be over for teardown.
    # We cannot wait on finished signal as _flash_animation may be already
    # removed when calling wait.
    qtbot.waitUntil(
        lambda: not hasattr(
            viewer.window._qt_viewer._welcome_widget, '_flash_animation'
        )
    )


@skip_on_win_ci
def test_changing_theme(make_napari_viewer):
    """Test changing the theme updates the full window."""
    viewer = make_napari_viewer(show=False)
    viewer.window._qt_viewer.set_welcome_visible(False)
    viewer.add_points(data=None)
    size = viewer.window._qt_viewer.size()
    viewer.window._qt_viewer.setFixedSize(size)

    assert viewer.theme == 'dark'
    screenshot_dark = viewer.screenshot(canvas_only=False, flash=False)

    viewer.theme = 'light'
    assert viewer.theme == 'light'
    screenshot_light = viewer.screenshot(canvas_only=False, flash=False)

    equal = (screenshot_dark == screenshot_light).min(-1)

    # more than 99.5% of the pixels have changed
    assert (np.count_nonzero(equal) / equal.size) < 0.05, 'Themes too similar'

    with pytest.raises(
        ValidationError, match="Theme 'nonexistent_theme' not found"
    ):
        viewer.theme = 'nonexistent_theme'


@pytest.mark.parametrize(('layer_class', 'data', 'ndim'), layer_test_data)
def test_roll_transpose_update(make_napari_viewer, layer_class, data, ndim):
    """Check that transpose and roll preserve correct transform sequence."""
    viewer = make_napari_viewer()

    np.random.seed(0)

    layer = add_layer_by_type(viewer, layer_class, data)

    # Set translations and scalings (match type of visual layer storing):
    transf_dict = {
        'translate': np.random.randint(0, 10, ndim).astype(np.float32),
        'scale': np.random.rand(ndim).astype(np.float32),
    }
    for k, val in transf_dict.items():
        setattr(layer, k, val)

    if layer_class in [layers.Image, layers.Labels]:
        transf_dict['translate'] -= transf_dict['scale'] / 2

    # Check consistency:
    check_view_transform_consistency(layer, viewer, transf_dict)

    # Roll dims and check again:
    viewer.dims.roll()
    check_view_transform_consistency(layer, viewer, transf_dict)

    # Transpose and check again:
    viewer.dims.transpose()
    check_view_transform_consistency(layer, viewer, transf_dict)


def test_toggling_axes(make_napari_viewer):
    """Test toggling axes."""
    viewer = make_napari_viewer()

    # Check axes are not visible
    assert not viewer.axes.visible

    # Make axes visible
    viewer.axes.visible = True
    assert viewer.axes.visible

    # Enter 3D rendering and check axes still visible
    viewer.dims.ndisplay = 3
    assert viewer.axes.visible

    # Make axes not visible
    viewer.axes.visible = False
    assert not viewer.axes.visible


def test_toggling_scale_bar(make_napari_viewer):
    """Test toggling scale bar."""
    viewer = make_napari_viewer()

    # Check scale bar is not visible
    assert not viewer.scale_bar.visible

    # Make scale bar visible
    viewer.scale_bar.visible = True
    assert viewer.scale_bar.visible

    # Enter 3D rendering and check scale bar is still visible
    viewer.dims.ndisplay = 3
    assert viewer.scale_bar.visible

    # Make scale bar not visible
    viewer.scale_bar.visible = False
    assert not viewer.scale_bar.visible


def test_removing_points_data(make_napari_viewer):
    viewer = make_napari_viewer()
    points = np.random.random((4, 2)) * 4

    pts_layer = viewer.add_points(points)
    pts_layer.data = np.zeros([0, 2])

    assert len(pts_layer.data) == 0


def test_deleting_points(make_napari_viewer):
    viewer = make_napari_viewer()
    points = np.random.random((4, 2)) * 4

    pts_layer = viewer.add_points(points)
    pts_layer.selected_data = {0}
    pts_layer.remove_selected()

    assert len(pts_layer.data) == 3


@skip_on_win_ci
@skip_local_popups
def test_custom_layer(make_napari_viewer):
    """Make sure that custom layers subclasses can be added to the viewer."""

    class NewLabels(layers.Labels):
        """'Empty' extension of napari Labels layer."""

    # Make a viewer and add the custom layer
    viewer = make_napari_viewer(show=True)
    viewer.add_layer(NewLabels(np.zeros((10, 10, 10), dtype=np.uint8)))


def test_emitting_data_doesnt_change_points_value(make_napari_viewer):
    """Test emitting data with no change doesn't change the layer _value."""
    viewer = make_napari_viewer()

    data = np.array([[0, 0], [10, 10], [20, 20]])
    layer = viewer.add_points(data, size=2)
    viewer.layers.selection.active = layer

    assert layer._value is None
    viewer.mouse_over_canvas = True
    viewer.cursor.position = tuple(layer.data[1])
    viewer._calc_status_from_cursor()
    assert layer._value == 1

    layer.events.data(value=layer.data)
    assert layer._value == 1


@pytest.mark.parametrize(('layer_class', 'data', 'ndim'), layer_test_data)
def test_emitting_data_doesnt_change_cursor_position(
    make_napari_viewer, layer_class, data, ndim
):
    """Test emitting data event from layer doesn't change cursor position"""
    viewer = make_napari_viewer()
    layer = layer_class(data)
    viewer.add_layer(layer)

    new_position = (5,) * ndim
    viewer.cursor.position = new_position
    layer.events.data(value=layer.data)

    assert viewer.cursor.position == new_position


@skip_local_popups
@skip_on_win_ci
def test_empty_shapes_dims(make_napari_viewer):
    """make sure an empty shapes layer can render in 3D"""
    viewer = make_napari_viewer(show=True)
    viewer.add_shapes(None)
    viewer.dims.ndisplay = 3


def test_current_viewer(make_napari_viewer):
    """Test that the viewer made last is the "current_viewer()" until another is activated"""
    # Make two DIFFERENT viewers
    viewer1: Viewer = make_napari_viewer()
    viewer2: Viewer = make_napari_viewer()
    assert viewer2 is not viewer1
    # Ensure one is returned by napari.current_viewer()
    from napari import current_viewer

    assert current_viewer() is viewer2
    assert current_viewer() is not viewer1

    viewer1.window.activate()

    assert current_viewer() is viewer1
    assert current_viewer() is not viewer2


@pytest.mark.parametrize('n_viewers', [1, 2, 3, 4])
def test_close_all(n_viewers, make_napari_viewer):
    """Test that close all closes multiple viewers"""

    # Make several viewers
    viewers = [make_napari_viewer() for _ in range(n_viewers)]
    last_viewer = viewers[-1]

    # Ensure the last viewer closes all
    assert last_viewer.close_all() == n_viewers


def test_reset_empty(make_napari_viewer):
    """
    Test that resetting an empty viewer doesn't crash
    https://github.com/napari/napari/issues/4867
    """
    viewer = make_napari_viewer()
    viewer.reset()


def test_reset_non_empty(make_napari_viewer):
    """
    Test that resetting a non-empty viewer doesn't crash
    https://github.com/napari/napari/issues/4867
    """
    viewer = make_napari_viewer()
    viewer.add_points([(0, 1), (2, 3)])
    viewer.reset()


def test_running_status_thread(make_napari_viewer, qtbot, monkeypatch):
    viewer = make_napari_viewer()
    settings = get_settings()
    start_mock, stop_mock = Mock(), Mock()
    monkeypatch.setattr(
        viewer.window._qt_window.status_thread, 'start', start_mock
    )
    monkeypatch.setattr(
        viewer.window._qt_window.status_thread, 'terminate', stop_mock
    )
    assert settings.appearance.update_status_based_on_layer
    settings.appearance.update_status_based_on_layer = False
    stop_mock.assert_called_once()
    start_mock.assert_not_called()
    settings.appearance.update_status_based_on_layer = True
    start_mock.assert_called_once()


def test_negative_translate(make_napari_viewer, qtbot):
    """Check that negative translation behaves as expected.

    See https://github.com/napari/napari/issues/7248
    """
    data = np.random.random((1, 3, 10, 12, 12))
    viewer = make_napari_viewer()
    _ = viewer.add_image(data, translate=(-1, 0, 0))
    assert viewer.dims.range[2].start == -1



"""Qt 'File' menu Actions."""

import sys
from pathlib import Path

from app_model.types import (
    Action,
    KeyCode,
    KeyMod,
    StandardKeyBinding,
    SubmenuItem,
)

from napari._app_model.constants import MenuGroup, MenuId
from napari._app_model.context import (
    LayerListContextKeys as LLCK,
    LayerListSelectionContextKeys as LLSCK,
)
from napari._qt.qt_main_window import Window
from napari._qt.qt_viewer import QtViewer
from napari._qt.widgets.qt_viewer_buttons import add_new_points, add_new_shapes
from napari.utils.translations import trans

# File submenus
FILE_SUBMENUS = [
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_NEW_LAYER,
            title=trans._('New Layer'),
            group=MenuGroup.NAVIGATION,
            order=0,
        ),
    ),
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_OPEN_WITH_PLUGIN,
            title=trans._('Open with Plugin'),
            group=MenuGroup.OPEN,
            order=99,
        ),
    ),
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_SAMPLES,
            title=trans._('Open Sample'),
            group=MenuGroup.OPEN,
            order=100,
        ),
    ),
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_IO_UTILITIES,
            title=trans._('IO Utilities'),
            group=MenuGroup.UTIL,
            order=101,
        ),
    ),
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_ACQUIRE,
            title=trans._('Acquire'),
            group=MenuGroup.UTIL,
            order=101,
        ),
    ),
]


# File actions


def new_labels(qt_viewer: QtViewer):
    viewer = qt_viewer.viewer
    viewer._new_labels()


def new_points(qt_viewer: QtViewer):
    viewer = qt_viewer.viewer
    add_new_points(viewer)


def new_shapes(qt_viewer: QtViewer):
    viewer = qt_viewer.viewer
    add_new_shapes(viewer)


def _open_files_with_plugin(qt_viewer: QtViewer):
    qt_viewer._open_files_dialog(choose_plugin=True)


def _open_files_as_stack_with_plugin(qt_viewer: QtViewer):
    qt_viewer._open_files_dialog_as_stack_dialog(choose_plugin=True)


def _open_folder_with_plugin(qt_viewer: QtViewer):
    qt_viewer._open_folder_dialog(choose_plugin=True)


def _save_selected_layers(qt_viewer: QtViewer):
    qt_viewer._save_layers_dialog(selected=True)


def _restart(window: Window):
    window._qt_window.restart()


def _close_window(window: Window):
    window._qt_window.close(quit_app=False, confirm_need=True)


def _close_app(window: Window):
    window._qt_window.close(quit_app=True, confirm_need=True)


Q_FILE_ACTIONS: list[Action] = [
    Action(
        id='napari.window.file.new_layer.new_labels',
        title=trans._('Labels'),
        callback=new_labels,
        menus=[{'id': MenuId.FILE_NEW_LAYER, 'group': MenuGroup.NAVIGATION}],
    ),
    Action(
        id='napari.window.file.new_layer.new_points',
        title=trans._('Points'),
        callback=new_points,
        menus=[{'id': MenuId.FILE_NEW_LAYER, 'group': MenuGroup.NAVIGATION}],
    ),
    Action(
        id='napari.window.file.new_layer.new_shapes',
        title=trans._('Shapes'),
        callback=new_shapes,
        menus=[{'id': MenuId.FILE_NEW_LAYER, 'group': MenuGroup.NAVIGATION}],
    ),
    Action(
        id='napari.window.file._image_from_clipboard',
        title=trans._('New Image from Clipboard'),
        callback=QtViewer._image_from_clipboard,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.NAVIGATION}],
        keybindings=[{'primary': KeyMod.CtrlCmd | KeyCode.KeyN}],
    ),
    Action(
        id='napari.window.file.open_files_dialog',
        title=trans._('Open File(s)...'),
        callback=QtViewer._open_files_dialog,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.OPEN}],
        keybindings=[StandardKeyBinding.Open],
    ),
    Action(
        id='napari.window.file.open_files_as_stack_dialog',
        title=trans._('Open Files as Stack...'),
        callback=QtViewer._open_files_dialog_as_stack_dialog,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.OPEN}],
        keybindings=[{'primary': KeyMod.CtrlCmd | KeyMod.Alt | KeyCode.KeyO}],
    ),
    Action(
        id='napari.window.file.open_folder_dialog',
        title=trans._('Open Folder...'),
        callback=QtViewer._open_folder_dialog,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.OPEN}],
        keybindings=[
            {'primary': KeyMod.CtrlCmd | KeyMod.Shift | KeyCode.KeyO}
        ],
    ),
    Action(
        id='napari.window.file._open_files_with_plugin',
        title=trans._('Open File(s)...'),
        callback=_open_files_with_plugin,
        menus=[{'id': MenuId.FILE_OPEN_WITH_PLUGIN, 'group': MenuGroup.OPEN}],
    ),
    Action(
        id='napari.window.file._open_files_as_stack_with_plugin',
        title=trans._('Open Files as Stack...'),
        callback=_open_files_as_stack_with_plugin,
        menus=[{'id': MenuId.FILE_OPEN_WITH_PLUGIN, 'group': MenuGroup.OPEN}],
    ),
    Action(
        id='napari.window.file._open_folder_with_plugin',
        title=trans._('Open Folder...'),
        callback=_open_folder_with_plugin,
        menus=[{'id': MenuId.FILE_OPEN_WITH_PLUGIN, 'group': MenuGroup.OPEN}],
    ),
    Action(
        id='napari.window.file.show_preferences_dialog',
        title=trans._('Preferences'),
        callback=Window._open_preferences_dialog,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.PREFERENCES}],
        # TODO: revert to `StandardKeyBinding.Preferences` after app-model>0.2.0
        keybindings=[{'primary': KeyMod.CtrlCmd | KeyCode.Comma}],
    ),
    # TODO:
    # If app-model supports a `kwargs` field (see:
    # https://github.com/pyapp-kit/app-model/issues/52)
    # it may allow registration of the same `id` when args are different and
    # we can re-use `DLG_SAVE_LAYERS` below.
    Action(
        id='napari.window.file.save_layers_dialog.selected',
        title=trans._('Save Selected Layers...'),
        callback=_save_selected_layers,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.SAVE}],
        keybindings=[StandardKeyBinding.Save],
        enablement=(LLSCK.num_selected_layers > 0),
    ),
    Action(
        id='napari.window.file.save_layers_dialog',
        title=trans._('Save All Layers...'),
        callback=QtViewer._save_layers_dialog,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.SAVE}],
        keybindings=[{'primary': KeyMod.CtrlCmd | KeyMod.Alt | KeyCode.KeyS}],
        enablement=(LLCK.num_layers > 0),
    ),
    Action(
        id='napari.window.file.save_canvas_screenshot_dialog',
        title=trans._('Save Screenshot...'),
        callback=QtViewer._screenshot_dialog,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.SAVE}],
        keybindings=[{'primary': KeyMod.Alt | KeyCode.KeyS}],
        status_tip=trans._(
            'Save screenshot of current display, default: .png'
        ),
    ),
    Action(
        id='napari.window.file.save_viewer_screenshot_dialog',
        title=trans._('Save Screenshot with Viewer...'),
        callback=Window._screenshot_dialog,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.SAVE}],
        keybindings=[{'primary': KeyMod.Alt | KeyMod.Shift | KeyCode.KeyS}],
        status_tip=trans._(
            'Save screenshot of current display, default: .png'
        ),
    ),
    Action(
        id='napari.window.file.copy_canvas_screenshot',
        title=trans._('Copy Screenshot to Clipboard'),
        callback=QtViewer.clipboard,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.SAVE}],
        keybindings=[{'primary': KeyMod.Alt | KeyCode.KeyC}],
        status_tip=trans._(
            'Copy screenshot of current display to the clipboard'
        ),
    ),
    Action(
        id='napari.window.file.copy_viewer_screenshot',
        title=trans._('Copy Screenshot with Viewer to Clipboard'),
        callback=Window.clipboard,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.SAVE}],
        keybindings=[{'primary': KeyMod.Alt | KeyMod.Shift | KeyCode.KeyC}],
        status_tip=trans._(
            'Copy screenshot of current display with the viewer to the clipboard'
        ),
    ),
    Action(
        id='napari.window.file.close_dialog',
        title=trans._('Close Window'),
        callback=_close_window,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.CLOSE}],
        keybindings=[StandardKeyBinding.Close],
    ),
    Action(
        id='napari.window.file.restart',
        title=trans._('Restart'),
        callback=_restart,
        menus=[
            {
                'id': MenuId.MENUBAR_FILE,
                'group': MenuGroup.CLOSE,
                'when': (
                    Path(sys.executable).parent / '.napari_is_bundled'
                ).exists(),
            }
        ],
    ),
    Action(
        id='napari.window.file.quit_dialog',
        title=trans._('Exit'),
        callback=_close_app,
        menus=[{'id': MenuId.MENUBAR_FILE, 'group': MenuGroup.CLOSE}],
        keybindings=[StandardKeyBinding.Quit],
    ),
]



from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QSpinBox


class QtSpinBox(QSpinBox):
    """Extends QSpinBox validate and stepBy methods in order
    to skip values in spin box."""

    prohibit = None

    def setProhibitValue(self, value: int):
        """Set value that should not be used in QSpinBox.

        Parameters
        ----------
        value : int
            Value to be excluded from QSpinBox.
        """
        self.prohibit = value

    def validate(self, value: str, pos: int):
        if value == str(self.prohibit):
            return QValidator.Invalid, value, pos
        return super().validate(value, pos)

    def stepBy(self, steps: int) -> None:
        if self.value() + steps == self.prohibit:
            steps *= 2
        return super().stepBy(steps)



import os
from pathlib import Path

from napari.settings import get_settings


def update_open_history(filename: str) -> None:
    """Updates open history of files in settings.

    Parameters
    ----------
    filename : str
        New file being added to open history.
    """
    settings = get_settings()
    folders = settings.application.open_history
    new_loc = os.path.dirname(filename)
    if new_loc in folders:
        folders.insert(0, folders.pop(folders.index(new_loc)))
    else:
        folders.insert(0, new_loc)

    folders = folders[0:10]
    settings.application.open_history = folders


def update_save_history(filename: str) -> None:
    """Updates save history of files in settings.

    Parameters
    ----------
    filename : str
        New file being added to save history.
    """
    settings = get_settings()
    folders = settings.application.save_history
    new_loc = os.path.dirname(filename)
    if new_loc in folders:
        folders.insert(0, folders.pop(folders.index(new_loc)))
    else:
        folders.insert(0, new_loc)

    folders = folders[0:10]
    settings.application.save_history = folders


def get_open_history() -> list[str]:
    """A helper for history handling."""
    settings = get_settings()
    folders = settings.application.open_history
    folders = [f for f in folders if os.path.isdir(f)]
    return folders or [str(Path.home())]


def get_save_history() -> list[str]:
    """A helper for history handling."""
    settings = get_settings()
    folders = settings.application.save_history
    folders = [f for f in folders if os.path.isdir(f)]
    return folders or [str(Path.home())]



"""Plugin related functions that require Qt.

Non-Qt plugin functions can be found in: `napari/plugins/_npe2.py`
"""

from __future__ import annotations

import inspect
from functools import partial
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    cast,
)

from app_model import Action
from app_model.types import SubmenuItem, ToggleRule
from magicgui.type_map._magicgui import MagicFactory
from magicgui.widgets import FunctionGui, Widget
from npe2 import plugin_manager as pm
from qtpy.QtWidgets import QWidget

from napari._app_model import get_app_model
from napari._app_model.constants import MenuGroup, MenuId
from napari._qt._qapp_model.injection._qproviders import (
    _provide_viewer_or_raise,
    _provide_window,
    _provide_window_or_raise,
)
from napari.errors.reader_errors import MultipleReaderError
from napari.plugins import menu_item_template, plugin_manager
from napari.plugins._npe2 import _when_group_order, get_widget_contribution
from napari.utils.events import Event
from napari.utils.translations import trans
from napari.viewer import Viewer

if TYPE_CHECKING:
    from npe2.manifest import PluginManifest
    from npe2.plugin_manager import PluginName
    from npe2.types import WidgetCreator

    from napari.qt import QtViewer


# TODO: This is a separate function from `_build_samples_submenu_actions` so it
# can be easily deleted once npe1 is no longer supported.
def _rebuild_npe1_samples_menu() -> None:  # pragma: no cover
    """Register submenu and actions for all npe1 plugins, clearing all first."""
    app = get_app_model()
    # Unregister all existing npe1 sample menu actions and submenus
    if unreg := plugin_manager._unreg_sample_submenus:
        unreg()
    if unreg := plugin_manager._unreg_sample_actions:
        unreg()

    sample_actions: list[Action] = []
    sample_submenus: list[Any] = []
    for plugin_name, samples in plugin_manager._sample_data.items():
        multiprovider = len(samples) > 1
        if multiprovider:
            submenu_id = f'napari/file/samples/{plugin_name}'
            submenu = (
                MenuId.FILE_SAMPLES,
                SubmenuItem(submenu=submenu_id, title=trans._(plugin_name)),
            )
            sample_submenus.append(submenu)
        else:
            submenu_id = MenuId.FILE_SAMPLES

        for sample_name, sample_dict in samples.items():
            _add_sample_partial = partial(
                _add_sample,
                plugin=plugin_name,
                sample=sample_name,
            )

            display_name = sample_dict['display_name'].replace('&', '&&')
            if multiprovider:
                title = display_name
            else:
                title = menu_item_template.format(plugin_name, display_name)

            action: Action = Action(
                id=f'{plugin_name}:{display_name}',
                title=title,
                menus=[{'id': submenu_id, 'group': MenuGroup.NAVIGATION}],
                callback=_add_sample_partial,
            )
            sample_actions.append(action)

    if sample_submenus:
        unreg_sample_submenus = app.menus.append_menu_items(sample_submenus)
        plugin_manager._unreg_sample_submenus = unreg_sample_submenus
    if sample_actions:
        unreg_sample_actions = app.register_actions(sample_actions)
        plugin_manager._unreg_sample_actions = unreg_sample_actions


# TODO: This should be deleted once npe1 is no longer supported.
def _toggle_or_get_widget_npe1(
    plugin: str,
    widget_name: str,
    name: str,
    hook_type: str,
) -> None:  # pragma: no cover
    """Toggle if widget already built otherwise return widget for npe1."""
    window = _provide_window_or_raise(
        msg='Note that widgets cannot be opened in headless mode.'
    )

    if window and (dock_widget := window._dock_widgets.get(name)):
        dock_widget.setVisible(not dock_widget.isVisible())
        return

    if hook_type == 'dock':
        window.add_plugin_dock_widget(plugin, widget_name)
    else:
        window._add_plugin_function_widget(plugin, widget_name)


def _rebuild_npe1_plugins_menu() -> None:
    """Register widget submenu and actions for all npe1 plugins, clearing all first."""
    app = get_app_model()

    # Unregister all existing npe1 plugin menu actions and submenus
    if unreg := plugin_manager._unreg_plugin_submenus:
        unreg()
    if unreg := plugin_manager._unreg_plugin_actions:
        unreg()

    widget_actions: list[Action] = []
    widget_submenus: list[Any] = []
    for hook_type, (plugin_name, widgets) in chain(
        plugin_manager.iter_widgets()
    ):
        multiprovider = len(widgets) > 1
        if multiprovider:
            submenu_id = f'napari/plugins/{plugin_name}'
            submenu = (
                MenuId.MENUBAR_PLUGINS,
                SubmenuItem(
                    submenu=submenu_id,
                    title=trans._(plugin_name),
                    group=MenuGroup.PLUGIN_MULTI_SUBMENU,
                ),
            )
            widget_submenus.append(submenu)
        else:
            submenu_id = MenuId.MENUBAR_PLUGINS

        for widget_name in widgets:
            full_name = menu_item_template.format(plugin_name, widget_name)
            title = widget_name if multiprovider else full_name

            _widget_callback = partial(
                _toggle_or_get_widget_npe1,
                plugin=plugin_name,
                widget_name=widget_name,
                name=full_name,
                hook_type=hook_type,
            )
            _get_current_dock_status_partial = partial(
                _get_current_dock_status,
                full_name=full_name,
            )
            action: Action = Action(
                id=f'{plugin_name}:{widget_name.replace("&", "&&")}',
                title=title.replace('&', '&&'),
                menus=[
                    {
                        'id': submenu_id,
                        'group': MenuGroup.PLUGIN_SINGLE_CONTRIBUTIONS,
                    }
                ],
                callback=_widget_callback,
                toggled=ToggleRule(
                    get_current=_get_current_dock_status_partial
                ),
            )
            widget_actions.append(action)

    if widget_submenus:
        unreg_plugin_submenus = app.menus.append_menu_items(widget_submenus)
        plugin_manager._unreg_plugin_submenus = unreg_plugin_submenus
    if widget_actions:
        unreg_plugin_actions = app.register_actions(widget_actions)
        plugin_manager._unreg_plugin_actions = unreg_plugin_actions


def _get_contrib_parent_menu(
    multiprovider: bool,
    parent_menu: MenuId,
    mf: PluginManifest,
    group: str | None = None,
) -> tuple[str, list[tuple[str, SubmenuItem]]]:
    """Get parent menu of plugin contribution (samples/widgets).

    If plugin provides multiple contributions, create a new submenu item.
    """
    submenu: list[tuple[str, SubmenuItem]] = []
    if multiprovider:
        submenu_id = f'{parent_menu}/{mf.name}'
        submenu = [
            (
                parent_menu,
                SubmenuItem(
                    submenu=submenu_id,
                    title=trans._(mf.display_name),
                    group=group,
                ),
            ),
        ]
    else:
        submenu_id = parent_menu
    return submenu_id, submenu


# Note `QtViewer` gets added to `injection_store.namespace` during
# `init_qactions` so does not need to be imported for type annotation resolution
def _add_sample(qt_viewer: QtViewer, plugin: str, sample: str) -> None:
    from napari._qt.dialogs.qt_reader_dialog import handle_gui_reading

    try:
        qt_viewer.viewer.open_sample(plugin, sample)
    except MultipleReaderError as e:
        handle_gui_reading(
            [str(p) for p in e.paths],
            qt_viewer,
            stack=False,
        )


def _build_samples_submenu_actions(
    mf: PluginManifest,
) -> tuple[list[tuple[str, SubmenuItem]], list[Action]]:
    """Build sample data submenu and actions for a single npe2 plugin manifest."""
    from napari._app_model.constants import MenuGroup, MenuId
    from napari.plugins import menu_item_template

    # If no sample data, return
    if not mf.contributions.sample_data:
        return [], []

    sample_data = mf.contributions.sample_data
    multiprovider = len(sample_data) > 1
    submenu_id, submenu = _get_contrib_parent_menu(
        multiprovider,
        MenuId.FILE_SAMPLES,
        mf,
    )

    sample_actions: list[Action] = []
    for sample in sample_data:
        _add_sample_partial = partial(
            _add_sample,
            plugin=mf.name,
            sample=sample.key,
        )

        if multiprovider:
            title = sample.display_name
        else:
            title = menu_item_template.format(
                mf.display_name, sample.display_name
            )
        # To display '&' instead of creating a shortcut
        title = title.replace('&', '&&')

        action: Action = Action(
            id=f'{mf.name}:{sample.key}',
            title=title,
            menus=[{'id': submenu_id, 'group': MenuGroup.NAVIGATION}],
            callback=_add_sample_partial,
        )
        sample_actions.append(action)
    return submenu, sample_actions


def _get_widget_viewer_param(
    widget_callable: WidgetCreator, widget_name: str
) -> str:
    """Get widget parameter name for `viewer` (if any) and check type."""
    if inspect.isclass(widget_callable) and issubclass(
        widget_callable,
        QWidget | Widget,
    ):
        widget_param = ''
        try:
            sig = inspect.signature(widget_callable.__init__)
        except ValueError:  # pragma: no cover
            # Inspection can fail when adding to bundled viewer as it thinks widget is
            # a builtin
            pass
        else:
            for param in sig.parameters.values():
                if param.name == 'napari_viewer' or param.annotation in (
                    'napari.viewer.Viewer',
                    Viewer,
                ):
                    widget_param = param.name
                    break

    # For magicgui type widget contributions, `Viewer` injection is done by
    # `magicgui.register_type`.
    elif isinstance(widget_callable, MagicFactory) or inspect.isfunction(
        widget_callable
    ):
        widget_param = ''
    else:
        raise TypeError(
            trans._(
                "'{widget}' must be `QtWidgets.QWidget` or `magicgui.widgets.Widget` subclass, `MagicFactory` instance or function. Please raise an issue in napari GitHub with the plugin and widget you were trying to use.",
                deferred=True,
                widget=widget_name,
            )
        )
    return widget_param


def _toggle_or_get_widget(
    plugin: str,
    widget_name: str,
    full_name: str,
) -> tuple[FunctionGui | QWidget | Widget, str] | None:
    """Toggle if widget already built otherwise return widget.

    Returned widget will be added to main window by a processor.
    Note for magicgui type widget contributions, `Viewer` injection is done by
    `magicgui.register_type` instead of a provider via annnotation.
    """
    viewer = _provide_viewer_or_raise(
        msg='Note that widgets cannot be opened in headless mode.',
    )

    window = viewer.window
    if window and (dock_widget := window._dock_widgets.get(full_name)):
        dock_widget.setVisible(not dock_widget.isVisible())
        return None

    # Get widget param name (if any) and check type
    widget_callable, _ = get_widget_contribution(plugin, widget_name)  # type: ignore [misc]
    widget_param = _get_widget_viewer_param(widget_callable, widget_name)

    kwargs = {}
    if widget_param:
        kwargs[widget_param] = viewer
    return widget_callable(**kwargs), full_name


def _get_current_dock_status(full_name: str) -> bool:
    window = _provide_window_or_raise(
        msg='Note that widgets cannot be opened in headless mode.',
    )
    if full_name in window._dock_widgets:
        return window._dock_widgets[full_name].isVisible()
    return False


def _build_widgets_submenu_actions(
    mf: PluginManifest,
) -> tuple[list[tuple[str, SubmenuItem]], list[Action]]:
    """Build widget submenu and actions for a single npe2 plugin manifest."""
    # If no widgets, return
    if not mf.contributions.widgets:
        return [], []

    # if this plugin declares any menu items, its actions should have the
    # plugin name.
    # TODO: update once plugin has self menus - they shouldn't exclude it
    # from the shorter name
    declares_menu_items = any(
        len(pm.instance()._command_menu_map[mf.name][command.id])
        for command in mf.contributions.commands or []
    )
    widgets = mf.contributions.widgets
    multiprovider = len(widgets) > 1
    default_submenu_id, default_submenu = _get_contrib_parent_menu(
        multiprovider,
        MenuId.MENUBAR_PLUGINS,
        mf,
        MenuGroup.PLUGIN_MULTI_SUBMENU,
    )
    needs_full_title = declares_menu_items or not multiprovider

    widget_actions: list[Action] = []
    for widget in widgets:
        full_name = menu_item_template.format(
            mf.display_name,
            widget.display_name,
        )

        _widget_callback = partial(
            _toggle_or_get_widget,
            plugin=mf.name,
            widget_name=widget.display_name,
            full_name=full_name,
        )
        _get_current_dock_status_partial = partial(
            _get_current_dock_status,
            full_name=full_name,
        )

        action_menus = [
            dict({'id': menu_key}, **_when_group_order(menu_item))
            for menu_key, menu_items in pm.instance()
            ._command_menu_map[mf.name][widget.command]
            .items()
            for menu_item in menu_items
        ] + [
            {
                'id': default_submenu_id,
                'group': MenuGroup.PLUGIN_SINGLE_CONTRIBUTIONS,
            }
        ]
        title = full_name if needs_full_title else widget.display_name
        # To display '&' instead of creating a shortcut
        title = title.replace('&', '&&')

        widget_actions.append(
            Action(
                id=f'{mf.name}:{widget.display_name}',
                title=title,
                callback=_widget_callback,
                menus=action_menus,
                toggled=ToggleRule(
                    get_current=_get_current_dock_status_partial
                ),
            )
        )
    return default_submenu, widget_actions


def _register_qt_actions(mf: PluginManifest) -> None:
    """Register samples and widget actions and submenus from a manifest.

    This is called when a plugin is registered or enabled and it adds the
    plugin's sample and widget actions and submenus to the app model registry.
    """
    app = get_app_model()
    samples_submenu, sample_actions = _build_samples_submenu_actions(mf)
    widgets_submenu, widget_actions = _build_widgets_submenu_actions(mf)

    context = pm.get_context(cast('PluginName', mf.name))
    actions = sample_actions + widget_actions
    if actions:
        context.register_disposable(app.register_actions(actions))
    submenus = samples_submenu + widgets_submenu
    if submenus:
        context.register_disposable(app.menus.append_menu_items(submenus))

    # Register dispose functions to remove plugin widgets from widget dictionary
    # `window._dock_widgets`
    if window := _provide_window():
        for widget in mf.contributions.widgets or ():
            widget_event = Event(type_name='', value=widget.display_name)

            def _remove_widget(event: Event = widget_event) -> None:
                window._remove_dock_widget(event)

            context.register_disposable(_remove_widget)



from napari.layers.labels._labels_constants import Mode
from napari.layers.labels._labels_utils import mouse_event_to_labels_coordinate
from napari.settings import get_settings


def draw(layer, event):
    """Draw with the currently selected label to a coordinate.

    This method have different behavior when draw is called
    with different labeling layer mode.

    In PAINT mode the cursor functions like a paint brush changing any
    pixels it brushes over to the current label. If the background label
    `0` is selected than any pixels will be changed to background and this
    tool functions like an eraser. The size and shape of the cursor can be
    adjusted in the properties widget.

    In FILL mode the cursor functions like a fill bucket replacing pixels
    of the label clicked on with the current label. It can either replace
    all pixels of that label or just those that are contiguous with the
    clicked on pixel. If the background label `0` is selected than any
    pixels will be changed to background and this tool functions like an
    eraser
    """

    # Do not allow drawing while adjusting the brush size with the mouse
    if layer.cursor == 'circle_frozen':
        return

    coordinates = mouse_event_to_labels_coordinate(layer, event)
    if layer._mode == Mode.ERASE:
        new_label = layer.colormap.background_value
    else:
        new_label = layer.selected_label

    # on press
    with layer.block_history():
        layer._draw(new_label, coordinates, coordinates)
        yield

        last_cursor_coord = coordinates
        # on move
        while event.type == 'mouse_move':
            coordinates = mouse_event_to_labels_coordinate(layer, event)
            if coordinates is not None or last_cursor_coord is not None:
                layer._draw(new_label, last_cursor_coord, coordinates)
            last_cursor_coord = coordinates
            yield


def pick(layer, event):
    """Change the selected label to the same as the region clicked."""
    # on press
    layer.selected_label = (
        layer.get_value(
            event.position,
            view_direction=event.view_direction,
            dims_displayed=event.dims_displayed,
            world=True,
        )
        or 0
    )


class BrushSizeOnMouseMove:
    """Enables changing the brush size by moving the mouse while holding down the specified modifiers

    When hold down specified modifiers and move the mouse,
    the callback will adjust the brush size based on the direction of the mouse movement.
    Moving the mouse right will increase the brush size, while moving it left will decrease it.
    The amount of change is proportional to the distance moved by the mouse.

    Parameters
    ----------
    min_brush_size : int
        The minimum brush size.

    """

    def __init__(self, min_brush_size: int = 1):
        self.min_brush_size = min_brush_size
        self.init_pos = None
        self.init_brush_size = None

        get_settings().application.events.brush_size_on_mouse_move_modifiers.connect(
            self._on_modifiers_change
        )
        self._on_modifiers_change()

    def __call__(self, layer, event):
        if all(modifier in event.modifiers for modifier in self.modifiers):
            pos = event.pos  # position in the canvas coordinates (x, y)

            if self.init_pos is None:
                self.init_pos = pos
                self.init_brush_size = layer.brush_size
                layer.cursor = 'circle_frozen'
            else:
                brush_size_delta = round(
                    (pos[0] - self.init_pos[0]) / event.camera_zoom
                )
                new_brush_size = self.init_brush_size + brush_size_delta

                bounded_brush_size = max(new_brush_size, self.min_brush_size)
                layer.brush_size = bounded_brush_size
        else:
            self.init_pos = None
            if layer.cursor == 'circle_frozen':
                layer.cursor = 'circle'

    def _on_modifiers_change(self):
        modifiers_setting = (
            get_settings().application.brush_size_on_mouse_move_modifiers
        )
        self.modifiers = modifiers_setting.value.split('+')
